//                 common.cpp
//
// lptr class.
// - lptr = long ptr.
// - each lptr consists of segm and offset values.
// - arithmetic is defined on these. Within Borg the seg/offset
//   values should be well defined. ie we do not test for seg1:offs1==seg2:offs2
//   where seg1!=seg2.
//   Borg addresses should be 'individual'.
//   May need to add conversions later, when looking at Dos exes in more detail.

#include <windows.h>
#include "common.h"

// predefined NULL pointer.
const lptr nlptr(0,0);

// lptr constructor with values.
lptr::lptr(word seg,dword off)
{ segm=seg;
  offs=off;
}

// lptr assignment, straightforward, as for constructor above.
void lptr::assign(word seg,dword off)
{ segm=seg;
  offs=off;
}

// lptr == operator, returns true if and only if both segs and offsets are equal
BOOL lptr::operator==(lptr loc2)
{ return ((segm==loc2.segm)&&(offs==loc2.offs));
}

// These operators are well defined given the constraints imposed on the seg:offs values
BOOL lptr::operator<=(lptr loc2)
{ if(segm!=loc2.segm) return (segm<=loc2.segm);
  return (offs<=loc2.offs);
}

BOOL lptr::operator>=(lptr loc2)
{ if(segm!=loc2.segm) return (segm>=loc2.segm);
  return (offs>=loc2.offs);
}

BOOL lptr::operator<(lptr loc2)
{ if(segm!=loc2.segm) return (segm<loc2.segm);
  return (offs<loc2.offs);
}

BOOL lptr::operator>(lptr loc2)
{ if(segm!=loc2.segm) return (segm>loc2.segm);
  return (offs>loc2.offs);
}

BOOL lptr::operator!=(lptr loc2)
{ return ((segm!=loc2.segm)||(offs!=loc2.offs));
}

// addition has been defined to allow a dword to be added to an offset.
// - this allows expressions like lptr r+1, etc to be used.
lptr lptr::operator+(dword offs2)
{ lptr tmp;
  tmp.segm=segm;
  tmp.offs=offs+offs2;
  return tmp;
}

#pragma warn -par
// increment lptr.
lptr lptr::operator++(int x)
{ offs++;
  return *this;
}
#pragma warn +par

// as for +.
lptr lptr::operator+=(dword offs2)
{ offs+=offs2;
  return *this;
}

// minus has been defined in a similar way to plus.
lptr lptr::operator-(dword offs2)
{ lptr tmp;
  tmp.segm=segm;
  tmp.offs=offs-offs2;
  return tmp;
}

// I have also defined lptr a - lptr b, as the difference in offsets,
// which is useful at times.
dword lptr::operator-(lptr loc2)
{ return offs-loc2.offs;
}



