//						disasm.cpp
//
// disasm class is the main workhorse.
// it builds up the disassembly list, and is called for any changes to this.

#include <windows.h>

#include "disasm.h"
#include "data.h"
#include "schedule.h"
#include "dasm.h"
#include "names.h"
#include "imports.h"
#include "exports.h"
#include "mainwind.h"
#include "debug.h"
#include "relocs.h"
#include "xref.h"

char hdr[] =";             Created by Borg Disassembler\r\n";
char hdr2[]=";                   written by Cronos\r\n";

int disasmcfunc(listitem i,listitem j);
void disasmdfunc(listitem i);

disasm::disasm()
{ setcomparefunc(disasmcfunc);
  setdelfunc(disasmdfunc);
  lastdis=NULL;
  subitem=dsmnull;
  stacktop=0;
  itables=0;
  jtables=0;
  irefs=0;
}

disasm::~disasm()
{
}

// sets up the instruction table to use.
// call after decided on the processor.
void disasm::dissettable(void)
{ int i=0;
  while((procnames[i].num!=0)&&(procnames[i].num!=options.processor)) i++;
  itable=procnames[i].tab;
}

// checks that not naming a location in the middle of an instruction
// returns TRUE if it can be named ok with no problem.
BOOL disasm::oktoname(lptr loc)
{ dsmitem check,*checker;
  check.addr=loc;
  check.type=dsmcode;
  // check table for the given address
  checker=(dsmitem *)find((listitem)&check);
  // NULL returned - must be ok.
  if(checker==NULL) return TRUE;
  // check bounds.
  if((checker->addr<loc)&&(checker->addr+checker->length>loc))
	 return FALSE;
  return TRUE;
}

// checkvalid - call when adding an instruction disassembly to the list.
// it checks that it is not being added in the middle of another instruction
// or over another instruction.
// - note, it also deletes any xrefs and names which get in the way!
// recoded in beta build 15. I wasn't happy with the previous version.
// - I think there may have been bugs when deleting several old xrefs/names at once.
// - new code is much cleaner.
BOOL disasm::checkvalid(dsmitem *newdsm)
{ dsmitem *lstdsm,*deldsm,fdsm/*,*d2*/;
  fdsm.addr=newdsm->addr;
  fdsm.type=dsmnull;
  lstdsm=(dsmitem *)find((listitem)&fdsm);
  if(lstdsm==NULL)return TRUE;
  // go through the disassembly items nearby and check for any overlaps.
  do
  { if((lstdsm->length)&&(lstdsm->addr.segm==newdsm->addr.segm))
	 { if((lstdsm->addr<=newdsm->addr)&&(lstdsm->addr+lstdsm->length>newdsm->addr))
		  return FALSE;
		if((newdsm->addr<=lstdsm->addr)&&(newdsm->addr+newdsm->length>lstdsm->addr))
		  return FALSE;
	 }
	 if(newdsm->addr+newdsm->length<=lstdsm->addr)break;
	 lstdsm=(dsmitem *)nextiterator();
  } while(lstdsm!=NULL);
  deldsm=(dsmitem *)find((listitem)&fdsm);
  // now go through them again, and this time delete any names/xrefs which get
  // in the way.
  do
  { if((!deldsm->length)&&(deldsm->addr.segm==newdsm->addr.segm))
	 { if((newdsm->addr<deldsm->addr)&&(newdsm->addr+newdsm->length>deldsm->addr))
		{ switch(deldsm->type)
		  { case dsmnameloc:
#ifdef DEBUG
				DebugMessage("Deleting Name : %s",deldsm->tptr);
#endif
				name.delname(deldsm->addr);
				break;
			 case dsmxref:
#ifdef DEBUG
				DebugMessage("Deleting Xref at : %04lx:%04lx",deldsm->addr.segm,deldsm->addr.offs);
#endif
				break;
			 default:
				break;
		  }
		  delfrom((listitem)deldsm);
		  deldsm=(dsmitem *)find((listitem)&fdsm);
		  if(deldsm==NULL)return TRUE;
		}
	 }
	 if(newdsm->addr+newdsm->length<=deldsm->addr)break;
	 deldsm=(dsmitem *)nextiterator();
  } while(lstdsm!=NULL);
  return TRUE;
}

// quick cleanstring function. This might be better off elsewhere, but where ?
void disasm::cleanstring(char *str)
{ unsigned int i;
  for(i=0;i<strlen(str);i++)
  { if(!isalnum(str[i]))str[i]='_';
  }
}

// sets argument override to decimal.
// - I'm still not quite convinced that this is the best way to go.
void disasm::disargoverdec(lptr loc)
{ dsmitem tester,*findit;
  tester.addr=loc;
  tester.type=dsmcode;
  findit=(dsmitem *)find((listitem)&tester);
  if(findit!=NULL)
  { if((findit->addr==loc)&&(findit->type==dsmcode))
	 {	findit->override=over_decimal;
		updatewindowifinrange(loc);
	 }
  }
}

// argument override reset.
void disasm::disargoverhex(lptr loc)
{ dsmitem tester,*findit;
  tester.addr=loc;
  tester.type=dsmcode;
  findit=(dsmitem *)find((listitem)&tester);
  if(findit!=NULL)
  { if((findit->addr==loc)&&(findit->type==dsmcode))
	 {	findit->override=over_null;
		updatewindowifinrange(loc);
	 }
  }
}

// nb - to do, remove any xref if changed to a different override.
void disasm::disargoveroffsetdseg(lptr loc)
{ dsmitem tester,*findit;
  lptr j;
  tester.addr=loc;
  tester.type=dsmcode;
  findit=(dsmitem *)find((listitem)&tester);
  if(findit!=NULL)
  { if((findit->addr==loc)&&(findit->type==dsmcode))
	 {	findit->override=over_dsoffset;
		updatewindowifinrange(loc);
		if((options.mode32)&&(findit->length>=4))
		{ j.assign(options.dseg,((dword *)(&findit->data[findit->length-4]))[0]);
		  xrefs.addxref(j,tester.addr);
		}
	 }
  }
}

// character override
void disasm::disargoverchar(lptr loc)
{ dsmitem tester,*findit;
  tester.addr=loc;
  tester.type=dsmcode;
  findit=(dsmitem *)find((listitem)&tester);
  if(findit!=NULL)
  { if((findit->addr==loc)&&(findit->type==dsmcode))
	 {	findit->override=over_char;
		updatewindowifinrange(loc);
	 }
  }
}

// disassembles a string at loc.
void disasm::disdatastring(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  dword maxlen,actuallen;
  char callit[GNAME_MAXLEN+1];
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  maxlen=dblock->size-(loc-dblock->addr);
  if(maxlen<2)return;
  actuallen=0;
  while((dblock->data+(loc-dblock->addr)+actuallen)[0])
  { actuallen++;
	 maxlen--;
	 if(!maxlen)return;
  }
  actuallen++;
  if(actuallen>0xffff)return; // tooo big
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmstr[0];
  newdsm->length=(word)actuallen;              // string length
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { callit[0]='s';
	 callit[1]='_';
	 if(actuallen>GNAME_MAXLEN-2)
	 { callit[GNAME_MAXLEN]=0;
		lstrcpyn(callit+2,(char *)newdsm->data,GNAME_MAXLEN-3);
	 }
	 else lstrcpy(callit+2,(char *)newdsm->data);
	 cleanstring(callit);
	 addto((listitem)newdsm);
	 name.addname(newdsm->addr,callit);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdataucstring(lptr loc)
{ // unicode c-style string.
  dsegitem *dblock;
  dsmitem *newdsm;
  dword maxlen,actuallen;
  unsigned int i;
  char callit[GNAME_MAXLEN+1];
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  maxlen=dblock->size-(loc-dblock->addr);
  if(maxlen<2)return;
  actuallen=0;
  while((dblock->data+(loc-dblock->addr)+actuallen)[0])
  { actuallen++;
	 maxlen--;
	 if(!maxlen)return;
	 actuallen++;
	 maxlen--;
	 if(!maxlen)return;
  }
  actuallen+=2;
  if(actuallen>0xffff)return; // tooo big
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmstr[3];
  newdsm->length=(word)actuallen;              // string length
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { callit[0]='s';
	 callit[1]='_';
	 i=2;
	 while(i<GNAME_MAXLEN-1)
	 { callit[i]=newdsm->data[(i-2)*2];
		i++;
		if(i*2>actuallen)break;
	 }
	 callit[i]=0;
	 cleanstring(callit);
	 addto((listitem)newdsm);
	 name.addname(newdsm->addr,callit);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdataupstring(lptr loc)
{ // unicode pascal style string.
  dsegitem *dblock;
  dsmitem *newdsm;
  word tlen;
  dword maxlen,actuallen;
  unsigned int i;
  char callit[GNAME_MAXLEN+1];
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  maxlen=dblock->size-(loc-dblock->addr);
  if(maxlen<2)return;
  maxlen-=2;
  actuallen=0;
  tlen=((word *)(dblock->data+(loc-dblock->addr)))[0];
  while(tlen)
  { actuallen++;
	 tlen--;
	 maxlen--;
	 if(!maxlen)return;
	 actuallen++;
	 maxlen--;
	 if(!maxlen)return;
  }
  actuallen+=2;
  if(actuallen>0xffff)return; // tooo big
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmstr[4];
  newdsm->length=(word)actuallen;              // string length
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { callit[0]='s';
	 callit[1]='_';
	 i=2;
	 while(i<GNAME_MAXLEN-1)
	 { callit[i]=newdsm->data[(i-1)*2];
		i++;
		if(i*2>actuallen)break;
	 }
	 callit[i]=0;
	 cleanstring(callit);
	 addto((listitem)newdsm);
	 name.addname(newdsm->addr,callit);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdatadosstring(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  dword maxlen,actuallen;
  char callit[GNAME_MAXLEN+1];
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  maxlen=dblock->size-(loc-dblock->addr);
  if(maxlen<2)return;
  actuallen=0;
  while((dblock->data+(loc-dblock->addr)+actuallen)[0]!='$')
  { actuallen++;
	 maxlen--;
	 if(!maxlen)return;
  }
  actuallen++;
  if(actuallen>0xffff)return; // tooo big
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmstr[2];
  newdsm->length=(word)actuallen;              // string length
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { callit[0]='s';
	 callit[1]='_';
	 callit[GNAME_MAXLEN]=0;
	 if(actuallen>GNAME_MAXLEN-2) lstrcpyn(callit+2,(char *)newdsm->data,GNAME_MAXLEN-3);
	 else lstrcpyn(callit+2,(char *)newdsm->data,actuallen);
	 cleanstring(callit);
	 addto((listitem)newdsm);
	 name.addname(newdsm->addr,callit);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdatageneralstring(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  dword maxlen,actuallen;
  char callit[GNAME_MAXLEN+1];
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  maxlen=dblock->size-(loc-dblock->addr);
  if(maxlen<2)return;
  actuallen=0;
  while(isprint((dblock->data+(loc-dblock->addr)+actuallen)[0]))
  { actuallen++;
	 maxlen--;
	 if(!maxlen)return;
  }
  actuallen++;
  if(actuallen>0xffff)return; // tooo big
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmstr[2];
  newdsm->length=(word)actuallen;              // string length
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { callit[0]='s';
	 callit[1]='_';
	 callit[GNAME_MAXLEN]=0;
	 if(actuallen>GNAME_MAXLEN-2) lstrcpyn(callit+2,(char *)newdsm->data,GNAME_MAXLEN-3);
	 else lstrcpyn(callit+2,(char *)newdsm->data,actuallen);
	 cleanstring(callit);
	 addto((listitem)newdsm);
	 name.addname(newdsm->addr,callit);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdatapstring(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  byte tlen;
  dword maxlen,actuallen;
  char callit[GNAME_MAXLEN+1];
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  maxlen=dblock->size-(loc-dblock->addr);
  if(maxlen<2)return;
  actuallen=0;
  tlen=(dblock->data+(loc-dblock->addr))[0];
  while(tlen)
  { actuallen++;
	 tlen--;
	 maxlen--;
	 if(!maxlen)return;
  }
  actuallen++;
  if(actuallen>0xffff)return; // tooo big
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmstr[1];
  newdsm->length=(word)actuallen;              // string length
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { callit[0]='s';
	 callit[1]='_';
	 if(actuallen>GNAME_MAXLEN-2)
	 { callit[GNAME_MAXLEN]=0;
		lstrcpyn(callit+2,(char *)newdsm->data+1,GNAME_MAXLEN-3);
	 }
	 else
	 { callit[actuallen+2]=0;
		lstrcpyn(callit+2,(char *)newdsm->data+1,actuallen);
	 }
	 cleanstring(callit);
	 addto((listitem)newdsm);
	 name.addname(newdsm->addr,callit);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdataword(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmword[0];
  newdsm->length=2;
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { addto((listitem)newdsm);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdatadword(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_null;
  newdsm->tptr=&asmdword[0];
  newdsm->length=4;
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { addto((listitem)newdsm);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

void disasm::disdatadsoffword(lptr loc)
{ dsegitem *dblock;
  dsmitem *newdsm;
  dblock=dta.findseg(loc);
  if(dblock==NULL)return;
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmcode;
  newdsm->override=over_dsoffset;
  newdsm->tptr=&asmdword[0];
  newdsm->length=4;
  newdsm->modrm=0;
  newdsm->data=dblock->data+(loc-dblock->addr);
  newdsm->mode32=options.mode32;
  newdsm->flags=0;
  if(checkvalid(newdsm))
  { addto((listitem)newdsm);
	 //check if need to update window.
	 updatewindowifinrange(loc);
  }
  else
	 delete newdsm;
}

// disassemble a block from this starting point.
void disasm::disblock(lptr loc)
{ byte ibuff[20];                     // ibuff is the disassembly buffer - m/c is moved here
												  // for disassembly - no checking of eof is then needed
												  // until after the code is identified.
  dsegitem *dblock;
  int i;
  BOOL doneblock;
  dsmitem thisblock,*tblock;
  dword aaddr;
  byte *mcode;
  asminstdata *codefound;
  int disasmcount;
  dblock=dta.findseg(loc);      // find segment item.
  disasmcount=0;
  if(dblock==NULL) return;
  if(dblock->typ==uninitdata)return;
  doneblock=FALSE;
  while(!doneblock)
  { disasmcount++;
	 // don't spend too long in here
	 if(disasmcount>50)
	 { scheduler.addtask(dis_code,priority_continuation,loc,NULL);
		return;
	 }
	 memset(ibuff,0,20);
	 // check if already done.
	 thisblock.addr=loc;
	 find((listitem)&thisblock);
	 tblock=(dsmitem *)nextiterator();
	 if(tblock!=NULL)
	 { while(!tblock->length)
		{ tblock=(dsmitem *)nextiterator();
		  if(tblock==NULL)break;
		}
	 }
	 if(tblock!=NULL)
		if((tblock->addr.segm==loc.segm)&&(tblock->addr<loc)&&(tblock->addr+tblock->length>loc))
		  doneblock=TRUE;
	 if(tblock!=NULL)
	 {	while(tblock->addr==loc)
		{ if(tblock->type!=dsmcomment)
		  { doneblock=TRUE;
			 break;
		  }
		  tblock=(dsmitem *)nextiterator();
		  if(tblock==NULL)break;
		}
	 }
	 // decode it.
	 if(doneblock)break;
	 aaddr=thisblock.addr-dblock->addr;
	 mcode=dblock->data+aaddr;
	 i=0;
	 while((i<15)&&(aaddr<dblock->size))
	 { ibuff[i]=mcode[i];
		i++;
		aaddr++;
	 }
	 tblock=decodeinst(ibuff,mcode,loc,TABLE_MAIN,options.mode32,0);
	 if(tblock!=NULL)
	 { loc.offs+=tblock->length;
		if((loc-dblock->addr)>dblock->size)
		{ doneblock=TRUE;
		  delfrom((listitem)tblock);
		}
	 }
	 else doneblock=TRUE;
	 if(doneblock)break;
	 // check if end (jmp,ret,etc)
	 codefound=(asminstdata *)tblock->tptr;
	 if((codefound->flags&FLAGS_JMP)||(codefound->flags&FLAGS_RET))
		doneblock=TRUE;
  }
}

dsmitem *disasm::decodeinst(byte *ibuff,byte *mcode,lptr loc,byte tabtype,BOOL omode32,int depth)
{ int tablenum=0;             // asmtable table number
  int instnum;                // instruction num in table
  asminstdata *insttab;
  dsmitem *newdsm;
  dword flgs;  					// inst flags
  argtype a1,a2,a3;           // inst args
  lptr j;                     // jump/call target
  byte *dta;
  byte length;
  gnameitem *imp;
  char *impname;
  BOOL righttable;
  byte cbyte,mbyte;
  BOOL fupbyte;
  cbyte=ibuff[0];
  if(tabtype==TABLE_EXT)cbyte=ibuff[1];
  if(tabtype==TABLE_EXT2)cbyte=ibuff[2];
  if((tabtype==TABLE_EXT2)&&(options.processor==PROC_Z80))cbyte=ibuff[3];
  while(itable[tablenum].table!=NULL)  // search tables
  { righttable=TRUE;
	 if((itable[tablenum].type!=tabtype)||(itable[tablenum].minlim>cbyte)||
		(itable[tablenum].maxlim<cbyte)) righttable=FALSE;
	 if(((tabtype==TABLE_EXT)||(tabtype==TABLE_EXT2))&&(ibuff[0]!=itable[tablenum].extnum))
		righttable=FALSE;
	 if((tabtype==TABLE_EXT2)&&(ibuff[1]!=itable[tablenum].extnum2))
		righttable=FALSE;
	 if(righttable)
	 { insttab=itable[tablenum].table;      // need to search this now
		instnum=0;
		mbyte=cbyte;
		fupbyte=FALSE;
		if(itable[tablenum].divisor)mbyte=(byte)(mbyte/itable[tablenum].divisor);
		if(itable[tablenum].mask)mbyte=mbyte&itable[tablenum].mask;
		else      // follow up byte encodings (KNI,AMD3DNOW)
		{ fupbyte=TRUE;
		  flgs=insttab[instnum].flags;
		  a1=insttab[instnum].arg1;
		  a2=insttab[instnum].arg2;
		  a3=insttab[instnum].arg3;
#pragma warn -sig
		  length=1+arglength(a1,mcode[1+itable[tablenum].modrmpos],mcode[2+itable[tablenum].modrmpos],flgs,TRUE)
					+arglength(a2,mcode[1+itable[tablenum].modrmpos],mcode[2+itable[tablenum].modrmpos],flgs,TRUE)
					+arglength(a3,mcode[1+itable[tablenum].modrmpos],mcode[2+itable[tablenum].modrmpos],flgs,TRUE)
					+itable[tablenum].modrmpos;
#pragma warn +sig
				// addition for table extensions where inst is part of modrm byte
		  if(((tabtype==TABLE_EXT)||(tabtype==TABLE_EXT2))&&(length==1)) length++;
		  mbyte=ibuff[length];
		}
		while((insttab[instnum].name!=NULL)||(insttab[instnum].instbyte)||(insttab[instnum].processor))
		{ if((omode32)&&(insttab[instnum].flags&FLAGS_OMODE16));
		  else if((!omode32)&&(insttab[instnum].flags&FLAGS_OMODE32));
		  else if((insttab[instnum].instbyte==mbyte)&&(insttab[instnum].processor&options.processor))
		  { // found it
			 if(insttab[instnum].name==NULL)
			 { if(tabtype==TABLE_MAIN)return decodeinst(ibuff,mcode,loc,TABLE_EXT,omode32,5);
				else return decodeinst(ibuff,mcode,loc,TABLE_EXT2,omode32,5);
			 }
			 else
			 { // interpret flags,etc
				flgs=insttab[instnum].flags;
				if((flgs&FLAGS_OPERPREFIX)&&(depth<5))
				{ newdsm=decodeinst(ibuff+1,mcode+1,loc+1,tabtype,!omode32,depth+1);
				  if(newdsm==NULL)return NULL;
				  newdsm->addr.offs--;
				  newdsm->length++;
				  newdsm->modrm++;
				  newdsm->data--;
				  return newdsm;
				}
				if((flgs&FLAGS_ADDRPREFIX)&&(depth<5))
				{ options.mode32=!options.mode32;
				  options.mode16=!options.mode16;
				  newdsm=decodeinst(ibuff+1,mcode+1,loc+1,tabtype,omode32,depth+1);
				  options.mode32=!options.mode32;
				  options.mode16=!options.mode16;
				  if(newdsm==NULL)return NULL;
				  newdsm->addr.offs--;
				  newdsm->length++;
				  newdsm->modrm++;
				  newdsm->data--;
				  newdsm->flags=newdsm->flags|FLAGS_ADDRPREFIX;
				  return newdsm;
				}
				if((flgs&FLAGS_SEGPREFIX)&&(depth<5))
				{ newdsm=decodeinst(ibuff+1,mcode+1,loc+1,tabtype,omode32,depth+1);
				  if(newdsm==NULL)return NULL;
				  newdsm->addr.offs--;
				  newdsm->length++;
				  newdsm->modrm++;
				  newdsm->data--;
				  newdsm->flags=newdsm->flags|FLAGS_SEGPREFIX;
				  return newdsm;
				}
				newdsm=new dsmitem;
				newdsm->addr=loc;
				newdsm->type=dsmcode;
				newdsm->override=over_null;
				newdsm->tptr=(void *)&insttab[instnum];
				newdsm->modrm=(byte)(1+itable[tablenum].modrmpos);
				newdsm->data=mcode;
				newdsm->mode32=omode32;
				if(flgs&FLAGS_16BIT) newdsm->mode32=FALSE;
				if(flgs&FLAGS_32BIT) newdsm->mode32=TRUE;
				newdsm->flags=flgs;
				a1=insttab[instnum].arg1;
				a2=insttab[instnum].arg2;
				a3=insttab[instnum].arg3;
#pragma warn -sig
				length=1+arglength(a1,mcode[1+itable[tablenum].modrmpos],mcode[2+itable[tablenum].modrmpos],flgs,newdsm->mode32)
					+arglength(a2,mcode[1+itable[tablenum].modrmpos],mcode[2+itable[tablenum].modrmpos],flgs,newdsm->mode32)
					+arglength(a3,mcode[1+itable[tablenum].modrmpos],mcode[2+itable[tablenum].modrmpos],flgs,newdsm->mode32)
					+itable[tablenum].modrmpos;
#pragma warn +sig
				// addition for table extensions where inst is part of modrm byte
				if(((tabtype==TABLE_EXT)||(tabtype==TABLE_EXT2))&&((length==1)||(options.processor==PROC_Z80)))
				  length++;
				if(options.processor==PROC_Z80)
				{ if(tabtype==TABLE_EXT2)length++;
				  if(flgs&FLAGS_INDEXREG)length++;
				}
				if(fupbyte)length++;
				newdsm->length=length;
				if(!checkvalid(newdsm))
				{ delete newdsm;
				  return NULL;
				}
				if(flgs&(FLAGS_JMP|FLAGS_CALL|FLAGS_CJMP))
				{ switch(a1)
				  { case ARG_RELIMM:
						j=loc;
						dta=mcode+length;
						if(options.mode32)
						{ dta-=4;
						  j+=((dword *)dta)[0]+length;
						}
						else
						{ dta-=2;
						  j+=(word)(((word *)dta)[0]+length);
						}
						scheduler.addtask(dis_code,priority_definitecode,j,NULL);
						xrefs.addxref(j,newdsm->addr);
						break;
					 case ARG_RELIMM8:
						j=loc;
						dta=mcode+length-1;
						if(options.mode32)
						{ if(dta[0]&0x80)
							 j+=(dword)(dta[0]+0xffffff00+length);
						  else
							 j+=(dword)(dta[0]+length);
						}
						else
						{ if(dta[0]&0x80)
							 j+=(word)(dta[0]+0xff00+length);
						  else
							 j+=(word)(dta[0]+length);
						}
						scheduler.addtask(dis_code,priority_definitecode,j,NULL);
						xrefs.addxref(j,newdsm->addr);
						break;
					 case ARG_FADDR:
						dta=mcode+length;
						if(options.mode32)
						{ dta-=6;
						  j.assign(((word *)(&dta[4]))[0],((dword *)(&dta[0]))[0]);
						}
						else
						{ dta-=4;
						  j.assign(((word *)(&dta[2]))[0],((word *)(&dta[0]))[0]);
						}
						scheduler.addtask(dis_code,priority_definitecode,j,NULL);
						xrefs.addxref(j,newdsm->addr);
						break;
					 case ARG_MODRM:
					 case ARG_MODRM_FPTR:
						if(options.mode32)
						{ if((newdsm->data[0]==0xff)&&(newdsm->data[1]==0x25)&&(tabtype==TABLE_EXT))
						  { j.assign(loc.segm,((dword *)(&newdsm->data[2]))[0]);
							 if(import.isname(j))
							 { imp=(gnameitem *)import.nextiterator();
								impname=new char[GNAME_MAXLEN+1];
								impname[0]='_';
								impname[GNAME_MAXLEN]=0;
								lstrcpyn(&impname[1],imp->name,GNAME_MAXLEN-2);
								//if(!name.isname(segm,offs))
								  scheduler.addtask(namecurloc,priority_nameloc,loc,impname);
							 }
						  }
						}
						scheduler.addtask(dis_jumptable,priority_definitecode,loc,NULL);
						break;
					 default:
						break;
				  }
				}
				switch(a1)
				{ case ARG_IMM32:
					 if(reloc.isreloc(loc+length-4))
					 {	newdsm->override=over_dsoffset;
						dta=mcode+length-4;
						j.assign(options.dseg,((dword *)(&dta[0]))[0]);
						xrefs.addxref(j,newdsm->addr);
					 }
					 break;
				  case ARG_IMM:
					 if(options.mode32)
					 { if(reloc.isreloc(loc+length-4))
						{ newdsm->override=over_dsoffset;
						  dta=mcode+length-4;
						  j.assign(options.dseg,((dword *)(&dta[0]))[0]);
						  xrefs.addxref(j,newdsm->addr);
						}
					 }
					 break;
				  default:
					 break;
				}
				switch(a2)
				{ case ARG_IMM32:
					 if(reloc.isreloc(loc+length-4))
					 {	newdsm->override=over_dsoffset;
						dta=mcode+length-4;
						j.assign(options.dseg,((dword *)(&dta[0]))[0]);
						xrefs.addxref(j,newdsm->addr);
					 }
					 break;
				  case ARG_IMM:
					 if(options.mode32)
					 { if(reloc.isreloc(loc+length-4))
						{ newdsm->override=over_dsoffset;
						  dta=mcode+length-4;
						  j.assign(options.dseg,((dword *)(&dta[0]))[0]);
						  xrefs.addxref(j,newdsm->addr);
						}
					 }
					 break;
				  default:
					 break;
				}
				addto((listitem)newdsm);
				//check if need to update window.
				updatewindowifinrange(loc);
				return newdsm;
			 }
		  }
		  instnum++;
		}
	 }
	 tablenum++;
  }
  return NULL;
}

void disasm::updatewindowifinrange(lptr loc)
{ if((loc>=curraddr)&&(loc<=outend))
	 scheduler.addtask(windowupdate,priority_window,nlptr,NULL);
}

// returns increase in length of instruction due to argtype
byte disasm::arglength(argtype a,byte modrmbyte,byte sibbyte,dword flgs,BOOL omode32)   //NB modrm to add
{ byte rm;
  switch(a)
  { case ARG_IMM:
		if(flgs&FLAGS_8BIT) return 1;
		if(!omode32) return 2;
		else return 4;
	 case ARG_NONEBYTE:
		return 1;
	 case ARG_RELIMM:
	 case ARG_MEMLOC:
		if(options.mode16) return 2;
		else return 4;
	 case ARG_RELIMM8:
	 case ARG_SIMM8:
	 case ARG_IMM8:
	 case ARG_IMM8_IND:
		return 1;
	 case ARG_IMM32:
		return 4;
	 case ARG_IMM16_A:
	 case ARG_IMM16:
	 case ARG_MEMLOC16:
		return 2;
	 case ARG_FADDR:
		if(options.mode16) return 4;
		else return 6;
	 case ARG_MODREG:
	 case ARG_MMXMODRM:
	 case ARG_XMMMODRM:
	 case ARG_MODRM8:
	 case ARG_MODRM16:
	 case ARG_MODRM_S:
	 case ARG_MODRMM512:
	 case ARG_MODRMQ:
	 case ARG_MODRM_SREAL:
	 case ARG_MODRM_PTR:
	 case ARG_MODRM_WORD:
	 case ARG_MODRM_SINT:
	 case ARG_MODRM_EREAL:
	 case ARG_MODRM_DREAL:
	 case ARG_MODRM_WINT:
	 case ARG_MODRM_LINT:
	 case ARG_MODRM_BCD:
	 case ARG_MODRM_FPTR:
	 case ARG_MODRM:
		rm=(byte)((modrmbyte&0xc0)>>6);
		switch(rm)
		{ case 0:
			 if(options.mode32)
			 { if((modrmbyte&0x07)==5)
				  return 5; // disp32
				if((modrmbyte&0x07)==4)
				{ if((sibbyte&0x07)==5)
					 return 6;
				  return 2; //sib byte - need to check if r=5 also.
				}
			 }
			 else if((modrmbyte&0x07)==6) return 3;
			 break;
		  case 1:
			 if(options.mode32)
			 { if((modrmbyte&0x07)==4)
				{ return 3; //sib byte
				}
			 }
			 return 2; // disp8
		  case 2:
			 if(options.mode32)
			 { if((modrmbyte&0x07)==4)
				  return 6; //sib byte
				return 5; // disp32
			 }
			 else return 3; // disp16
		  case 3:
			 return 1;
		}
		return 1;
	 default:
		break;
  }
  return 0;
}

void disasm::outargs(dsmitem *inst,argtype a)
{ byte *dta,modrm,sib;
  // rm extended to word. build 15. re M Ogden and VC++ warnings.
  word rm;
  argtype a1,a2;
  dword targetd;
  dword targetw;
  lptr loc;
  int i;
  BOOL pname;
  byte pbyte; // prefix byte.
  if(inst->flags&FLAGS_SEGPREFIX)
  { pbyte=inst->data[0];
	 if((pbyte==0x66)||(pbyte==0x67))pbyte=inst->data[1];
	 if((pbyte==0x66)||(pbyte==0x67))pbyte=inst->data[2];
  }
  if(inst->flags&FLAGS_ADDRPREFIX)options.mode32=!options.mode32;
  switch(a)
  { case ARG_REG_AX:
		if(inst->mode32)LastPrintBuff("eax");
		else LastPrintBuff("ax");
		break;
	 case ARG_REG_BX:
		if(inst->mode32)LastPrintBuff("ebx");
		else LastPrintBuff("bx");
		break;
	 case ARG_REG_CX:
		if(inst->mode32)LastPrintBuff("ecx");
		else LastPrintBuff("cx");
		break;
	 case ARG_REG_DX:
		if(inst->mode32)LastPrintBuff("edx");
		else LastPrintBuff("dx");
		break;
	 case ARG_16REG_DX:
		LastPrintBuff("dx");
		break;
	 case ARG_REG_SP:
		if(inst->mode32)LastPrintBuff("esp");
		else LastPrintBuff("sp");
		break;
	 case ARG_REG_BP:
		if(inst->mode32)LastPrintBuff("ebp");
		else LastPrintBuff("bp");
		break;
	 case ARG_REG_SI:
		if(inst->mode32)LastPrintBuff("esi");
		else LastPrintBuff("si");
		break;
	 case ARG_REG_DI:
		if(inst->mode32)LastPrintBuff("edi");
		else LastPrintBuff("di");
		break;
	 case ARG_REG_AL:
		LastPrintBuff("al");
		break;
	 case ARG_REG_AH:
		LastPrintBuff("ah");
		break;
	 case ARG_REG_BL:
		LastPrintBuff("bl");
		break;
	 case ARG_REG_BH:
		LastPrintBuff("bh");
		break;
	 case ARG_REG_CL:
		LastPrintBuff("cl");
		break;
	 case ARG_REG_CH:
		LastPrintBuff("ch");
		break;
	 case ARG_REG_DL:
		LastPrintBuff("dl");
		break;
	 case ARG_REG_DH:
		LastPrintBuff("dh");
		break;
	 case ARG_REG_ST0:
		LastPrintBuff("st(0)");
		break;
	 case ARG_REG_ES:
		LastPrintBuff("es");
		break;
	 case ARG_REG_CS:
		LastPrintBuff("cs");
		break;
	 case ARG_REG_DS:
		LastPrintBuff("ds");
		break;
	 case ARG_REG_SS:
		LastPrintBuff("ss");
		break;
	 case ARG_REG_FS:
		LastPrintBuff("fs");
		break;
	 case ARG_REG_GS:
		LastPrintBuff("gs");
		break;
	 case ARG_REG_A:
		LastPrintBuff("a");
		break;
	 case ARG_REG_B:
		LastPrintBuff("b");
		break;
	 case ARG_REG_C:
		LastPrintBuff("c");
		break;
	 case ARG_REG_D:
		LastPrintBuff("d");
		break;
	 case ARG_REG_E:
		LastPrintBuff("e");
		break;
	 case ARG_REG_H:
		LastPrintBuff("h");
		break;
	 case ARG_REG_L:
		LastPrintBuff("l");
		break;
	 case ARG_REG_I:
		LastPrintBuff("i");
		break;
	 case ARG_REG_R:
		LastPrintBuff("r");
		break;
	 case ARG_REG_HL_IND:
		LastPrintBuff("(hl)");
		break;
	 case ARG_REG_BC:
		LastPrintBuff("bc");
		break;
	 case ARG_REG_DE:
		LastPrintBuff("de");
		break;
	 case ARG_REG_HL:
		LastPrintBuff("hl");
		break;
	 case ARG_REG_BC_IND:
		LastPrintBuff("(bc)");
		break;
	 case ARG_REG_DE_IND:
		LastPrintBuff("(de)");
		break;
	 case ARG_REG_SP_IND:
		LastPrintBuff("(sp)");
		break;
	 case ARG_REG_IX:
		LastPrintBuff("ix");
		break;
	 case ARG_REG_IX_IND:
		LastPrintBuff("(ix");
		if(inst->flags&FLAGS_INDEXREG)
		{ dta=inst->data+2;
		  LastPrintBuff("+%02lxh",(word)(dta[0]));
		}
		LastPrintBuff(")");
		break;
	 case ARG_REG_IY:
		LastPrintBuff("iy");
		break;
	 case ARG_REG_IY_IND:
		LastPrintBuff("(iy");
		if(inst->flags&FLAGS_INDEXREG)
		{ dta=inst->data+2;
		  LastPrintBuff("+%02lxh",(word)(dta[0]));
		}
		LastPrintBuff(")");
		break;
	 case ARG_REG_C_IND:
		LastPrintBuff("(c)");
		break;
	 case ARG_REG_AF:
		LastPrintBuff("af");
		break;
	 case ARG_REG_AF2:
		LastPrintBuff("af\'");
		break;
	 case ARG_IMM:
		dta=inst->data+inst->length;
		if(inst->mode32)
		{ dta-=4;
		  switch(inst->override)
		  { case over_decimal:
				LastPrintBuff("%02lu",((dword *)dta)[0]);
				break;
			 case over_char:
				{ LastPrintBuff("\"");
				  for(i=3;i>=0;i--)
					 if(dta[i]) LastPrintBuff("%c",dta[i]);
				  LastPrintBuff("\"");
				}
				break;
			 case over_dsoffset:
				loc.assign(inst->addr.segm,((dword *)dta)[0]);
				LastPrintBuff("offset ");
				if(name.isname(loc)) name.printname(loc);
				else LastPrintBuff("%02lxh",((dword *)dta)[0]);
				break;
			 default:
				LastPrintBuff("%02lxh",((dword *)dta)[0]);
				break;
		  }
		}
		else
		{ dta-=2;
		  if(inst->override==over_decimal)
			 LastPrintBuff("%02lu",((word *)dta)[0]);
		  else if(inst->override==over_char)
		  { LastPrintBuff("\"");
			 for(i=1;i>=0;i--)
				if(dta[i]) LastPrintBuff("%c",dta[i]);
			 LastPrintBuff("\"");
		  }
		  else
			 LastPrintBuff("%02xh",((word *)dta)[0]);
		}
		break;
	 case ARG_IMM32:
		dta=inst->data+inst->length-4;
		switch(inst->override)
		{ case over_decimal:
			 LastPrintBuff("%02lu",((dword *)dta)[0]);
			 break;
		  case over_char:
			 LastPrintBuff("\"");
			 for(i=3;i>=0;i--)
				if(dta[i]) LastPrintBuff("%c",dta[i]);
			 LastPrintBuff("\"");
			 break;
		  case over_dsoffset:
			 loc.assign(inst->addr.segm,((dword *)dta)[0]);
			 LastPrintBuff("offset ");
			 if(name.isname(loc)) name.printname(loc);
			 else LastPrintBuff("%02lxh",loc.offs);
			 break;
		  default:
			 LastPrintBuff("%02lxh",((dword *)dta)[0]);
			 break;
		}
		break;
	 case ARG_STRING:
		LastPrintBuff("\"%s\"",inst->data);
		break;
	 case ARG_PSTRING:
		dta=inst->data;
		rm=dta[0];
		if(!rm)break;
		dta++;
		LastPrintBuff("\"");
		while(rm)
		{ LastPrintBuff("%c",dta[0]);
		  dta++;
		  rm--;
		}
		LastPrintBuff("\"");
		break;
	 case ARG_DOSSTRING:
		dta=inst->data;
		rm=inst->length;
		rm--;
		if(!rm)break;
		LastPrintBuff("\"");
		while(rm)
		{ LastPrintBuff("%c",dta[0]);
		  dta++;
		  rm--;
		}
		LastPrintBuff("\"");
		break;
	 case ARG_CUNICODESTRING:
		dta=inst->data;
		rm=inst->length;
		rm-=(byte)2;
		if(!rm)break;
		LastPrintBuff("\"");
		while(rm)
		{ LastPrintBuff("%c",dta[0]);
		  dta+=2;
		  rm-=(byte)2;
		}
		LastPrintBuff("\"");
		break;
	 case ARG_PUNICODESTRING:
		dta=inst->data+2;
		rm=inst->length;
		rm-=(byte)2;
		if(!rm)break;
		LastPrintBuff("\"");
		while(rm)
		{ LastPrintBuff("%c",dta[0]);
		  dta+=2;
		  rm-=(byte)2;
		}
		LastPrintBuff("\"");
		break;
	 case ARG_MEMLOC:
		if(inst->flags&FLAGS_SEGPREFIX)outprefix(pbyte);
		dta=inst->data+inst->length;
		if(options.mode32)
		{ dta-=4;
		  loc.assign(inst->addr.segm,((dword *)dta)[0]);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("[%02lxh]",loc.offs);
		}
		else
		{ dta-=2;
		  loc.assign(inst->addr.segm,((word *)dta)[0]);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("[%02xh]",loc.offs);
		}
		break;
	 case ARG_MEMLOC16:
		if(inst->flags&FLAGS_SEGPREFIX)outprefix(pbyte);
		dta=inst->data+inst->length-2;
		if(options.processor==PROC_Z80)
		{ loc.assign(inst->addr.segm,((dword *)dta)[0]);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("(%02xh)",loc.offs);
		}
		else
		{ loc.assign(inst->addr.segm,((word *)dta)[0]);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("[%02xh]",loc.offs);
		}
		break;
	 case ARG_SIMM8:
		dta=inst->data+inst->length-1;
		if(inst->override==over_char)
		{ LastPrintBuff("\"");
		  for(i=0;i>=0;i--)
			 if(dta[i]) LastPrintBuff("%c",dta[i]);
		  LastPrintBuff("\"");
		}
		else if(dta[0]&0x80)
		{ if(inst->override==over_decimal)
			 LastPrintBuff("%02lu",(word)(0x100-dta[0]));
		  else
			 LastPrintBuff("-%02xh",(word)(0x100-dta[0]));
		}
		else
		{ if(inst->override==over_decimal)
			 LastPrintBuff("%02lu",(word)(dta[0]));
		  else
			 LastPrintBuff("%02xh",(word)(dta[0]));
		}
		break;
	 case ARG_IMM8:
		dta=inst->data+inst->length-1;
		if(inst->override==over_decimal)
		  LastPrintBuff("%02lu",(word)(dta[0]));
		else if(inst->override==over_char)
		{ LastPrintBuff("\"");
		  for(i=0;i>=0;i--)
			 if(dta[i]) LastPrintBuff("%c",dta[i]);
		  LastPrintBuff("\"");
		}
		else
		  LastPrintBuff("%02xh",(word)(dta[0]));
		break;
	 case ARG_IMM8_IND:
		dta=inst->data+inst->length-1;
		LastPrintBuff("(%02xh)",(word)(dta[0]));
		break;
	 case ARG_IMM16:
		dta=inst->data+inst->length-2;
		if(inst->override==over_decimal)
		  LastPrintBuff("%02lu",((word *)dta)[0]);
		else if(inst->override==over_char)
		{ LastPrintBuff("\"");
		  for(i=1;i>=0;i--)
			 if(dta[i]) LastPrintBuff("%c",dta[i]);
		  LastPrintBuff("\"");
		}
		else
		  LastPrintBuff("%02xh",((word *)dta)[0]);
		break;
	 case ARG_IMM16_A:
		dta=inst->data+inst->length-3;
		LastPrintBuff("%02xh",((word *)dta)[0]);
		break;
	 case ARG_RELIMM8:
		dta=inst->data+inst->length-1;
		if(inst->mode32)
		{ if(dta[0]&0x80)
			 targetd=(dword)(dta[0]+0xffffff00+inst->addr.offs+inst->length);
		  else
			 targetd=(dword)(dta[0]+inst->addr.offs+inst->length);
		  loc.assign(inst->addr.segm,targetd);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("%02lxh",loc.offs);
		}
		else
		{ if(dta[0]&0x80)
			 targetw=(word)(dta[0]+0xff00+inst->addr.offs+inst->length);
		  else
			 targetw=(word)(dta[0]+inst->addr.offs+inst->length);
		  loc.assign(inst->addr.segm,targetw);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("%02xh",loc.offs);
		}
		break;
	 case ARG_RELIMM:
		dta=inst->data+inst->length;
		if(inst->mode32)
		{ dta-=4;
		  targetd=((dword *)dta)[0]+inst->addr.offs+inst->length;
		  loc.assign(inst->addr.segm,targetd);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("%02lxh",loc.offs);
		}
		else
		{ dta-=2;
		  targetw=(word)(((word *)dta)[0]+inst->addr.offs+inst->length);
		  loc.assign(inst->addr.segm,targetw);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("%02xh",loc.offs);
		}
		break;
	 case ARG_REG:
		dta=inst->data+inst->modrm;
		if(options.processor==PROC_Z80) LastPrintBuff(regzascii[dta[0]&0x07]);
		else if(((asminstdata *)(inst->tptr))->flags&FLAGS_8BIT) LastPrintBuff(reg8ascii[(dta[0]>>3)&0x07]);
		else if(inst->mode32) LastPrintBuff(reg32ascii[(dta[0]>>3)&0x07]);
		else LastPrintBuff(reg16ascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_MREG:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regmascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_XREG:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regxascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_FREG:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regfascii[dta[0]&0x07]);
		break;
	 case ARG_SREG:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regsascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_CREG:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regcascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_DREG:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regdascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_TREG:
	 case ARG_TREG_67:
		dta=inst->data+inst->modrm;
		LastPrintBuff(regtascii[(dta[0]>>3)&0x07]);
		break;
	 case ARG_MODREG:
	 case ARG_MMXMODRM:
	 case ARG_XMMMODRM:
	 case ARG_MODRM8:
	 case ARG_MODRM16:
	 case ARG_MODRM_S:
	 case ARG_MODRMM512:
	 case ARG_MODRMQ:
	 case ARG_MODRM_SREAL:
	 case ARG_MODRM_PTR:
	 case ARG_MODRM_WORD:
	 case ARG_MODRM_BCD:
	 case ARG_MODRM_SINT:
	 case ARG_MODRM_EREAL:
	 case ARG_MODRM_DREAL:
	 case ARG_MODRM_WINT:
	 case ARG_MODRM_LINT:
	 case ARG_MODRM_FPTR:
	 case ARG_MODRM:
		dta=inst->data+inst->modrm;
		rm=(byte)((dta[0]&0xc0)>>6);
		modrm=(byte)(dta[0]&0x07);
		a1=((asminstdata *)inst->tptr)->arg1;
		a2=((asminstdata *)inst->tptr)->arg2;
		if((a1==ARG_IMM)||(a2==ARG_IMM)||(a1==ARG_IMM8)||(a2==ARG_IMM8)||(a2==ARG_NONE)
		  ||(a1==ARG_SIMM8)||(a2==ARG_SIMM8))
		{ if(rm<3)
		  { switch(a)
			 { case ARG_MODRM8:
				  LastPrintBuff("byte ptr ");
				  break;
				case ARG_MODRM16:
				case ARG_MODRM_WORD:
				  LastPrintBuff("word ptr ");
				  break;
				case ARG_MMXMODRM:
				case ARG_XMMMODRM:
				  LastPrintBuff("dword ptr ");
				  break;
				case ARG_MODRMQ:
				  LastPrintBuff("qword ptr ");
				  break;
				case ARG_MODRM_S:
				  LastPrintBuff("6bytes ptr ");
				  break;
				case ARG_MODRM_SREAL:
				  LastPrintBuff("single-real ptr ");
				  break;
				case ARG_MODRM_BCD:
				  LastPrintBuff("packed-bcd ptr ");
				  break;
				case ARG_MODRM_SINT:
				  LastPrintBuff("short-int ptr ");
				  break;
				case ARG_MODRM_WINT:
				  LastPrintBuff("word-int ptr ");
				  break;
				case ARG_MODRM_LINT:
				  LastPrintBuff("long-int ptr ");
				  break;
				case ARG_MODRMM512:
				  LastPrintBuff("m512byte ptr ");
				  break;
				case ARG_MODRM_EREAL:
				  LastPrintBuff("extended-real ptr ");
				  break;
				case ARG_MODRM_DREAL:
				  LastPrintBuff("double-real ptr ");
				  break;
				case ARG_MODRM:
				  if(inst->flags&FLAGS_8BIT) LastPrintBuff("byte ptr ");
				  else if(inst->mode32) LastPrintBuff("dword ptr ");
				  else LastPrintBuff("word ptr ");
				  break;
				default:
				  break;
			 }
		  }
		}
		switch(rm)
		{ case 0:
			 if(inst->flags&FLAGS_SEGPREFIX)outprefix(pbyte);
			 if(options.mode32)
			 { if(modrm==5)
				{ loc.assign(inst->addr.segm,((dword *)(&dta[1]))[0]);
				  if(name.isname(loc)) name.printname(loc);
				  else if(import.isname(loc)) import.printname(loc);
						  else if(export.isname(loc)) export.printname(loc);
				  else LastPrintBuff("[%02lxh]",loc.offs);
				}
				else if(modrm==4)        // case 4=sib
				{ sib=dta[1];
				  if((sib&0x07)==5) // disp32
				  { loc.assign(inst->addr.segm,((dword *)(&dta[2]))[0]);
					 if(name.isname(loc))
					 {	name.printname(loc);
						pname=TRUE;
					 }
					 else if(import.isname(loc))
					 {	import.printname(loc);
						pname=TRUE;
					 }
					 else if(export.isname(loc))
					 {	export.printname(loc);
						pname=TRUE;
					 }
					 else
					 {	LastPrintBuff("[%02lxh",loc.offs);
						pname=FALSE;
					 }
				  }
				  else
				  { LastPrintBuff("[%s",reg32ascii[sib&0x07]);
					 pname=FALSE;
				  }
				  if(((sib>>3)&0x07)==4) // no scaled index reg
				  { if(!pname)LastPrintBuff("]");
				  }
				  else
				  { if(pname) LastPrintBuff("[%s",reg32ascii[(sib>>3)&0x07]);
					 else LastPrintBuff("+%s",reg32ascii[(sib>>3)&0x07]);
					 switch(sib>>6)
					 { case 0:
						  LastPrintBuff("]");
						  break;
						case 1:
						  LastPrintBuff("*2]");
						  break;
						case 2:
						  LastPrintBuff("*4]");
						  break;
						case 3:
						  LastPrintBuff("*8]");
						  break;
					 }
				  }
				}
				else
				  LastPrintBuff("[%s]",reg32ascii[dta[0]&0x07]);
			 }
			 else
			 { if(modrm==6)
				{ loc.assign(inst->addr.segm,((word *)(&dta[1]))[0]);
				  if(name.isname(loc)) name.printname(loc);
				  else if(import.isname(loc)) import.printname(loc);
				  else if(export.isname(loc)) export.printname(loc);
				  else LastPrintBuff("[%02xh]",loc.offs);
				}
				else
				  LastPrintBuff("[%s]",regind16ascii[dta[0]&0x07]);
			 }
			 break;
		  case 1:
			 if(inst->flags&FLAGS_SEGPREFIX)outprefix(pbyte);
			 if(options.mode32)
			 {	if(modrm==4)        // case 4=sib
				{ sib=dta[1];
				  if(dta[2]&0x80)
					 LastPrintBuff("[%s-%02xh",reg32ascii[dta[1]&0x07],0x100-dta[2]);
				  else
					 LastPrintBuff("[%s+%02xh",reg32ascii[dta[1]&0x07],dta[2]);
				  if(((sib>>3)&0x07)==4) // no scaled index reg
					 LastPrintBuff("]");
				  else
				  { LastPrintBuff("+%s",reg32ascii[(sib>>3)&0x07]);
					 switch(sib>>6)
					 { case 0:
						  LastPrintBuff("]");
						  break;
						case 1:
						  LastPrintBuff("*2]");
						  break;
						case 2:
						  LastPrintBuff("*4]");
						  break;
						case 3:
						  LastPrintBuff("*8]");
						  break;
					 }
				  }
				}
				else if(dta[1]&0x80)
				  LastPrintBuff("[%s-%02xh]",reg32ascii[dta[0]&0x07],0x100-dta[1]);
				else
				  LastPrintBuff("[%s+%02xh]",reg32ascii[dta[0]&0x07],dta[1]);
			 }
			 else
			 { if(dta[1]&0x80)
				  LastPrintBuff("[%s-%02xh]",regind16ascii[dta[0]&0x07],0x100-dta[1]);
				else
				  LastPrintBuff("[%s+%02xh]",regind16ascii[dta[0]&0x07],dta[1]);
			 }
			 break;
		  case 2:
			 if(inst->flags&FLAGS_SEGPREFIX)outprefix(pbyte);
			 if(options.mode32)
			 { loc.assign(inst->addr.segm,((dword *)(&dta[1]))[0]);
				if(modrm==4)        // case 4=sib
				{ sib=dta[1];
				  loc.assign(inst->addr.segm,((dword *)(&dta[2]))[0]);
				  if(name.isname(loc))
				  { name.printname(loc);
					 LastPrintBuff("[%s",reg32ascii[sib&0x07]);
				  }
				  else if(import.isname(loc))
				  { import.printname(loc);
					 LastPrintBuff("[%s",reg32ascii[sib&0x07]);
				  }
				  else if(export.isname(loc))
				  { export.printname(loc);
					 LastPrintBuff("[%s",reg32ascii[sib&0x07]);
				  }
				  else if(dta[5]&0x80)
					 LastPrintBuff("[%s-%02lxh",reg32ascii[sib&0x07],0-((dword *)(&dta[2]))[0]);
				  else
					 LastPrintBuff("[%s+%02lxh",reg32ascii[sib&0x07],((dword *)(&dta[2]))[0]);
				  if(((sib>>3)&0x07)==4) // no scaled index reg
					 LastPrintBuff("]");
				  else
				  { LastPrintBuff("+%s",reg32ascii[(sib>>3)&0x07]);
					 switch(sib>>6)
					 { case 0:
						  LastPrintBuff("]");
						  break;
						case 1:
						  LastPrintBuff("*2]");
						  break;
						case 2:
						  LastPrintBuff("*4]");
						  break;
						case 3:
						  LastPrintBuff("*8]");
						  break;
					 }
				  }
				}
				else if(name.isname(loc))
				{ name.printname(loc);
				  LastPrintBuff("[%s]",reg32ascii[dta[0]&0x07]);
				}
				else if(import.isname(loc))
				{ import.printname(loc);
				  LastPrintBuff("[%s]",reg32ascii[dta[0]&0x07]);
				}
				else if(export.isname(loc))
				{ export.printname(loc);
				  LastPrintBuff("[%s]",reg32ascii[dta[0]&0x07]);
				}
				else if(dta[4]&0x80)
				  LastPrintBuff("[%s-%02lxh]",reg32ascii[dta[0]&0x07],0-((dword *)(&dta[1]))[0]);
				else
				  LastPrintBuff("[%s+%02lxh]",reg32ascii[dta[0]&0x07],((dword *)(&dta[1]))[0]);
			 }
			 else
			 { loc.assign(inst->addr.segm,((word *)(&dta[1]))[0]);
				if(name.isname(loc))
				{ name.printname(loc);
				  LastPrintBuff("[%s]",regind16ascii[dta[0]&0x07]);
				}
				else if(import.isname(loc))
				{ import.printname(loc);
				  LastPrintBuff("[%s]",regind16ascii[dta[0]&0x07]);
				}
				else if(export.isname(loc))
				{ export.printname(loc);
				  LastPrintBuff("[%s]",regind16ascii[dta[0]&0x07]);
				}
				else if(dta[2]&0x80)
				  LastPrintBuff("[%s-%02xh]",regind16ascii[dta[0]&0x07],0x10000-((word *)(&dta[1]))[0]);
				else
				  LastPrintBuff("[%s+%02xh]",regind16ascii[dta[0]&0x07],((word *)(&dta[1]))[0]);
			 }
			 break;
		  case 3:
			 if(a==ARG_MMXMODRM) LastPrintBuff(regmascii[dta[0]&0x07]);
			 else if(a==ARG_XMMMODRM) LastPrintBuff(regxascii[dta[0]&0x07]);
			 else if((((asminstdata *)(inst->tptr))->flags&FLAGS_8BIT)||(a==ARG_MODRM8)) LastPrintBuff(reg8ascii[dta[0]&0x07]);
			 else if((inst->mode32)&&(a!=ARG_MODRM16)) LastPrintBuff(reg32ascii[dta[0]&0x07]);
			 else LastPrintBuff(reg16ascii[dta[0]&0x07]);
			 break;
		}
		break;
	 case ARG_IMM_1:
		if(inst->override==over_decimal)
		  LastPrintBuff("1");
		else
		  LastPrintBuff("1h");
		break;
	 case ARG_FADDR:
		dta=inst->data+inst->length;
		if(options.mode32)
		{ dta-=6;
		  loc.assign(((word *)(&dta[4]))[0],((dword *)(&dta[0]))[0]);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("%04x:%08lxh",loc.segm,loc.offs);
		}
		else
		{ dta-=4;
		  loc.assign(((word *)(&dta[2]))[0],((word *)(&dta[0]))[0]);
		  if(name.isname(loc)) name.printname(loc);
		  else if(import.isname(loc)) import.printname(loc);
		  else if(export.isname(loc)) export.printname(loc);
		  else LastPrintBuff("%04x:%04xh",loc.segm,loc.offs);
		}
		break;
	 case ARG_BIT:
		dta=inst->data+inst->length-1;
		LastPrintBuff("%x",(dta[0]>>3)&7);
		break;
	 case ARG_NONE:
		if(inst->flags&FLAGS_SEGPREFIX)outprefix(pbyte);
		break;
	 case ARG_NONEBYTE:
	 default:
		break;
  }
  if(inst->flags&FLAGS_ADDRPREFIX)options.mode32=!options.mode32;
}

// this functions job to output a window
// of disassembly to the buffer.
void disasm::updatewindow(void)
{ dsmitem titem,*tblock;
  lptr outhere;
  int i;
  // find current position.
  titem.addr=curraddr;
  titem.type=subitem;
  find((listitem)&titem);
  tblock=(dsmitem *)nextiterator();
  // now tblock= current position, or previous one.
  // check if in middle of instruction
  if(tblock!=NULL)
  { if((tblock->addr.segm==curraddr.segm)&&(tblock->addr<curraddr)&&
	  (tblock->addr+tblock->length>curraddr))
	  curraddr.offs=tblock->addr.offs;
  }
  // check if previous one. get next
  if(tblock!=NULL)
  { if((tblock->addr<curraddr)||
		((tblock->addr==curraddr)&&(tblock->type<subitem)))
		tblock=(dsmitem *)nextiterator();
  }
  // tblock is now top of the page.
  outhere=curraddr;
  ClearBuff();                        // start again
  for(i=0;i<buffer_lines;i++)
  { if(tblock!=NULL)
	 { if(outhere==tblock->addr)
		{ switch(tblock->type)
		  { case dsmcode:
				outinst(tblock,TRUE);
				break;
			 case dsmnameloc:
				outnameloc(tblock);
				break;
			 case dsmxref:
				PrintBuff(";");
				LastPrintBuffEpos(COMMENTPOS);
				LastPrintBuff("XREFS First: ");
				xrefs.printfirst(tblock->addr);
				break;
			 default:
				outcomment(tblock);
				break;
		  }
		  outhere+=tblock->length;
		  tblock=(dsmitem *)nextiterator();
		}
		else
		{ outdb(&outhere,TRUE);
		  outhere++;
		}
	 }
	 else
	 { outdb(&outhere,TRUE);
		outhere++;
	 }
	 // check if gone beyond seg, get next seg.
	 if(dta.beyondseg(outhere))
		dta.nextseg(&outhere);
	 if(!outhere.segm) break;
	 else outend=outhere;
  }
  DoneBuff();                         // mark as done
  InvalidateRect(mainwindow,NULL,TRUE);
}

void disasm::scroller(dword amount)
{ dsmitem titem,*tblock;
  RECT r;
  // simple move on screen
  if((amount==1)&&(usersel<=nScreenRows-3))
  { usersel++;
	 r.left=0;
	 r.right=rrr;
	 r.top=(usersel-1)*cyc;
	 r.bottom=(usersel+1)*cyc;
	 InvalidateRect(mainwindow,&r,TRUE);
	 return;
  }
  // simple move on screen
  if((amount==(dword)(-1))&&(usersel))
  { usersel--;
	 r.left=0;
	 r.right=rrr;
	 r.top=usersel*cyc;
	 r.bottom=(usersel+2)*cyc;
	 InvalidateRect(mainwindow,&r,TRUE);
	 return;
  }
  // find top of page
  titem.addr=curraddr;
  titem.type=subitem;
  tblock=(dsmitem *)find((listitem)&titem);
  // tblock is now top of page or previous item.
  // if moving up find previous inst
  if(amount<100)tblock=(dsmitem *)nextiterator();
  if(amount>100)
	 if(tblock!=NULL)
		if((tblock->addr==titem.addr)&&(tblock->type==subitem))
			 tblock=(dsmitem *)lastiterator();
  // move up - tblock=previous inst.
  // if moving down, check if tblock=previous item.
  if((tblock!=NULL)&&(amount<100))
  { if((tblock->addr<curraddr)||
		((tblock->addr==curraddr)&&(tblock->type<subitem)))
		tblock=(dsmitem *)nextiterator();
  }
  // moving down - tblock=current top.
  if(amount<100)
	 while(amount)
	 { if(usersel<=nScreenRows-3)usersel++;
		else
		{ if(tblock!=NULL)
		  { if(curraddr==tblock->addr)
			 { curraddr+=tblock->length;
				tblock=(dsmitem *)nextiterator();
				if(tblock!=NULL)
				{ if(curraddr==tblock->addr)subitem=tblock->type;
				  else subitem=dsmcode;
				}
				else subitem=dsmcode;
			 }
			 else
			 { subitem=dsmnull;
				curraddr++;
			 }
		  }
		  else
		  { subitem=dsmnull;
			 curraddr++;
		  }
		  // check if gone beyond seg, get next seg.
		  if(dta.beyondseg(curraddr))
		  { dta.nextseg(&curraddr);
			 subitem=dsmnull;
		  }
		  if(!curraddr.segm) break;
		  titem.addr=curraddr;
		}
		amount--;
	 }
  else
  { while(amount)
	 { if(usersel) usersel--;
		else
		{ if(tblock!=NULL)
		  { if(curraddr==tblock->addr+tblock->length)
			 { curraddr.offs-=tblock->length;
				subitem=tblock->type;
				tblock=(dsmitem *)lastiterator();
			 }
			 else
			 { subitem=dsmcode;
				curraddr.offs--;
			 }
		  }
		  else
		  { subitem=dsmnull;
			 curraddr.offs--;
		  }
		  // check if gone beyond seg, get next seg.
		  if(dta.beyondseg(curraddr))
			 dta.lastseg(&curraddr);
		  if(!curraddr.segm) break;
		  titem.addr=curraddr;
		}
		amount++;
	 }
  }
  if(!curraddr.segm) curraddr=titem.addr;
  updatewindow();
}

//compare function to order in list
int disasmcfunc(listitem i,listitem j)
{ if(((dsmitem *)i)->addr==((dsmitem *)j)->addr)
  { if(((dsmitem *)i)->type==((dsmitem *)j)->type)
		return 0;
	 else if(((dsmitem *)i)->type>((dsmitem *)j)->type)
		return 1;
	 return -1;
  }
  if(((dsmitem *)i)->addr>((dsmitem *)j)->addr)
	 return 1;
  return -1;
}

//deletion function for list
void disasmdfunc(listitem i)
{ // bugfix by Mark Ogden - added dsmnameloc
  if((((dsmitem *)i)->type!=dsmcode)&&(((dsmitem *)i)->type!=dsmnameloc))
	 if(((dsmitem *)i)->tptr!=NULL) delete ((dsmitem *)i)->tptr;
  delete (dsmitem *)i;
}

void disasm::setcuraddr(lptr loc)
{ dsmitem titem,*tblock;
  curraddr=loc;
  titem.addr=curraddr;
  titem.type=dsmnull;
  find((listitem)&titem);
  tblock=(dsmitem *)nextiterator();
  if(tblock!=NULL)
	 if((tblock->addr.segm<loc.segm)||((tblock->addr<loc)&&(tblock->addr+tblock->length<=loc)))
		tblock=(dsmitem *)nextiterator();
  if(tblock!=NULL)
  { subitem=tblock->type;
	 while(!tblock->length)
	 { tblock=(dsmitem *)nextiterator();
		if(tblock==NULL)break;
	 }
  }
  else subitem=dsmnull;
  if(tblock!=NULL)
  { if((tblock->addr.segm==loc.segm)&&(tblock->addr<loc)&&(tblock->addr+tblock->length>loc))
		curraddr=tblock->addr;
  }
  outend=loc+50;
  usersel=0;
  userselonscreen=TRUE;
  updatewindow();
}

void disasm::setpos(int ypos)
{ if(usersel!=ypos/cyc)
  { usersel=ypos/cyc;
	 InvalidateRect(mainwindow,NULL,TRUE);
  }
}

void disasm::outinst(dsmitem *inst,BOOL printaddrs)
{ dsegitem *dblock;
  int i;
  if(options.mode32)
  { if(printaddrs)PrintBuff("%04lx:%08lx",inst->addr.segm,inst->addr.offs);
	 else PrintBuff("");
	 LastPrintBuffEpos(BYTEPOS);
  }
  else
  { if(printaddrs)PrintBuff("%04lx:%04lx",inst->addr.segm,inst->addr.offs);
	 else PrintBuff("");
	 LastPrintBuffEpos(BYTEPOS-4);
  }
  dblock=dta.findseg(inst->addr);      // find segment item.
  switch(inst->type)
  { case dsmcode:
		i=inst->length;
		if(printaddrs)
		{ while(i)
		  { if(dblock!=NULL)
			 {	if(dblock->typ==uninitdata) LastPrintBuff("??");
				else LastPrintBuff("%02x",inst->data[inst->length-i]);
			 }
			 else LastPrintBuff("%02x",inst->data[inst->length-i]);
			 i--;
			 if(((i+8)<inst->length)&&(inst->length>10))
			 { LastPrintBuff("..");
				break;
			 }
		  }
		}
		if(options.mode32)
		  LastPrintBuffEpos(ASMPOS+4);
		else
		  LastPrintBuffEpos(ASMPOS);
		LastPrintBuff(((asminstdata *)inst->tptr)->name);
		LastPrintBuff(" ");
		if(options.mode32)
		  LastPrintBuffEpos(ARGPOS+4);
		else
		  LastPrintBuffEpos(ARGPOS);
		if(dblock!=NULL)
		{ if(dblock->typ==uninitdata) LastPrintBuff("??");
		  else
		  { outargs(inst,((asminstdata *)inst->tptr)->arg1);
			 if(((asminstdata *)inst->tptr)->arg2!=ARG_NONE)
			 { LastPrintBuff(", ");
				outargs(inst,((asminstdata *)inst->tptr)->arg2);
			 }
			 if(((asminstdata *)inst->tptr)->arg3!=ARG_NONE)
			 { LastPrintBuff(", ");
				outargs(inst,((asminstdata *)inst->tptr)->arg3);
			 }
		  }
		}
		else
		{ outargs(inst,((asminstdata *)inst->tptr)->arg1);
		  if(((asminstdata *)inst->tptr)->arg2!=ARG_NONE)
		  { LastPrintBuff(", ");
			 outargs(inst,((asminstdata *)inst->tptr)->arg2);
		  }
		  if(((asminstdata *)inst->tptr)->arg3!=ARG_NONE)
		  { LastPrintBuff(", ");
			 outargs(inst,((asminstdata *)inst->tptr)->arg3);
		  }
		}
		break;
	 default:
		break;
  }
}

void disasm::outdb(lptr *lp,BOOL printaddrs)
{ dsegitem *dblock;
  dword aaddr;
  byte *mcode;
  if(options.mode32)
  { if(printaddrs)PrintBuff("%04lx:%08lx",lp->segm,lp->offs);
	 else PrintBuff("");
	 LastPrintBuffEpos(BYTEPOS);
  }
  else
  { if(printaddrs)PrintBuff("%04lx:%04lx",lp->segm,lp->offs);
	 else PrintBuff("");
	 LastPrintBuffEpos(BYTEPOS-4);
  }
  dblock=dta.findseg(*lp);      // find segment item.
  if(dblock==NULL)
  { if(printaddrs)LastPrintBuff("??");
  }
  else if(dblock->typ==uninitdata)
  { if(printaddrs)LastPrintBuff("??");
  }
  else
  { aaddr=(*lp-dblock->addr);
	 mcode=dblock->data+aaddr;
	 if(printaddrs)LastPrintBuff("%02x",mcode[0]);
  }
  if(options.mode32)
	 LastPrintBuffEpos(ASMPOS+4);
  else
	 LastPrintBuffEpos(ASMPOS);
  LastPrintBuff("db");
  if(options.mode32)
	 LastPrintBuffEpos(ARGPOS+4);
  else
	 LastPrintBuffEpos(ARGPOS);
  if(dblock==NULL)
	 LastPrintBuff("??");
  else if(dblock->typ==uninitdata)
  { if(printaddrs)LastPrintBuff("??");
  }
  else
  { LastPrintBuff("%02x",mcode[0]);
	 if(isprint(mcode[0]))
	 { LastPrintBuffEpos(COMMENTPOS);
		LastPrintBuff(";\'%c\'",mcode[0]);
	 }
  }
}

void disasm::outprefix(byte prefixbyte)
{ switch(prefixbyte)
  { case 0x2e:
		LastPrintBuff("cs:");
		break;
	 case 0x36:
		LastPrintBuff("ss:");
		break;
	 case 0x3e:
		LastPrintBuff("ds:");
		break;
	 case 0x26:
		LastPrintBuff("es:");
		break;
	 case 0x64:
		LastPrintBuff("fs:");
		break;
	 case 0x65:
		LastPrintBuff("gs:");
		break;
	 default:
		LastPrintBuff("err:");
		break;
  }
}

void disasm::makedword(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_datadword,priority_userrequest,outhere,NULL);
}

void disasm::makeword(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_dataword,priority_userrequest,outhere,NULL);
}

void disasm::makestring(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_datastring,priority_userrequest,outhere,NULL);
}

void disasm::pascalstring(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_datapstring,priority_userrequest,outhere,NULL);
}

void disasm::ucstring(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_dataucstring,priority_userrequest,outhere,NULL);
}

void disasm::upstring(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_dataupstring,priority_userrequest,outhere,NULL);
}

void disasm::dosstring(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_datadosstring,priority_userrequest,outhere,NULL);
}

void disasm::generalstring(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_datageneralstring,priority_userrequest,outhere,NULL);
}

void disasm::makecode(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_code,priority_userrequest,outhere,NULL);
}

void disasm::argoverdec(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_argoverdec,priority_userrequest,outhere,NULL);
}

void disasm::argoverhex(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_argoverhex,priority_userrequest,outhere,NULL);
}

void disasm::argoveroffsetdseg(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_argoveroffsetdseg,priority_userrequest,outhere,NULL);
}

void disasm::argoverchar(void)
{ lptr outhere;
  findcurrentaddr(&outhere);
  scheduler.addtask(dis_argoverchar,priority_userrequest,outhere,NULL);
}

void disasm::undefineline(void)
{ dsmitem *tblock;
  lptr outhere;
  tblock=findcurrentline();
  if(tblock!=NULL)
  { outhere=tblock->addr;
  }
  if(tblock!=NULL)
  { while(!tblock->length)
	 { tblock=(dsmitem *)nextiterator();
		if(tblock==NULL)break;
	 }
  }
  if(tblock!=NULL)
  { if(outhere==tblock->addr)
		delfrom((listitem)tblock);
  }
  updatewindow();
}

void disasm::undefinelines(void)
{ dsmitem *tblock;
  lptr outhere;
  int i;
  tblock=findcurrentline();
  if(tblock!=NULL)
  { outhere=tblock->addr;
  }
  for(i=0;i<10;i++)
  { if(tblock!=NULL)
	 { while(!tblock->length)
		{ tblock=(dsmitem *)nextiterator();
		  if(tblock==NULL)break;
		}
	 }
	 if(tblock!=NULL)
	 { if(outhere==tblock->addr)
		{ outhere+=tblock->length;
		  delfrom((listitem)tblock);
		  tblock=(dsmitem *)nextiterator();
		}
		else break;
	 }
  }
  updatewindow();
}

void disasm::jumpback(void)
{ lptr outhere;
  outhere=nlptr;
  popcallstack(&outhere);
  if(outhere.segm)setcuraddr(outhere);
  updatewindow();
}

void disasm::jumpto(void)
{ dsmitem *tblock;
  lptr outhere;
  byte *data;
  BOOL madejump;
  tblock=findcurrentline();
  madejump=FALSE;
  if(tblock!=NULL)
  { switch(((asminstdata *)tblock->tptr)->arg1)
	 { case ARG_FADDR:
		  data=tblock->data+tblock->length;
		  if(tblock->mode32)
		  { data-=6;
			 outhere.assign(((word *)(&data[4]))[0],((dword *)(&data[0]))[0]);
		  }
		  else
		  { data-=4;
			 outhere.assign(((word *)(&data[2]))[0],((word *)(&data[0]))[0]);
		  }
		  if(dta.findseg(outhere)!=NULL)
		  { pushcallstack(curraddr);
			 setcuraddr(outhere);
			 updatewindow();
			 madejump=TRUE;
		  }
		  break;
		case ARG_IMM32:
		  if(tblock->override!=over_dsoffset)break;
		  data=tblock->data+tblock->length;
		  data-=4;
		  outhere.assign(options.dseg,((dword *)(&data[0]))[0]);
		  if(dta.findseg(outhere)!=NULL)
		  { pushcallstack(curraddr);
			 setcuraddr(outhere);
			 updatewindow();
			 madejump=TRUE;
		  }
		  break;
		case ARG_IMM:
		  if(tblock->override!=over_dsoffset)break;
		  if(options.mode32)
		  { data=tblock->data+tblock->length;
			 data-=4;
			 outhere.assign(options.dseg,((dword *)(&data[0]))[0]);
			 if(dta.findseg(outhere)!=NULL)
			 { pushcallstack(curraddr);
				setcuraddr(outhere);
				updatewindow();
				madejump=TRUE;
			 }
		  }
		  break;
		case ARG_RELIMM:
		  data=tblock->data+tblock->length;
		  outhere=tblock->addr;
		  if(tblock->mode32)
		  { data-=4;
			 outhere+=((dword *)data)[0]+tblock->length;
		  }
		  else
		  { data-=2;
			 outhere+=(word)(((word *)data)[0]+tblock->length);
		  }
		  if(dta.findseg(outhere)!=NULL)
		  { pushcallstack(curraddr);
			 setcuraddr(outhere);
			 updatewindow();
			 madejump=TRUE;
		  }
		  break;
		case ARG_RELIMM8:
		  data=tblock->data+tblock->length-1;
		  outhere=tblock->addr;
		  if(tblock->mode32)
		  { if(data[0]&0x80)
				outhere+=(dword)(data[0]+0xffffff00+tblock->length);
			 else
				outhere+=(dword)(data[0]+tblock->length);
		  }
		  else
		  { if(data[0]&0x80)
				outhere+=(word)(data[0]+0xff00+tblock->length);
			 else
				outhere+=(word)(data[0]+tblock->length);
		  }
		  if(dta.findseg(outhere)!=NULL)
		  { pushcallstack(curraddr);
			 setcuraddr(outhere);
			 updatewindow();
			 madejump=TRUE;
		  }
		  break;
		default:
		  break;
	 }
	 if(!madejump)
	 { switch(((asminstdata *)tblock->tptr)->arg2)
		{ case ARG_IMM32:
			 if(tblock->override!=over_dsoffset)break;
			 data=tblock->data+tblock->length;
			 data-=4;
			 outhere.assign(options.dseg,((dword *)(&data[0]))[0]);
			 if(dta.findseg(outhere)!=NULL)
			 { pushcallstack(curraddr);
				setcuraddr(outhere);
				updatewindow();
			 }
			 break;
		  case ARG_IMM:
			 if(tblock->override!=over_dsoffset)break;
			 if(options.mode32)
			 { data=tblock->data+tblock->length;
				data-=4;
				outhere.assign(options.dseg,((dword *)(&data[0]))[0]);
				if(dta.findseg(outhere)!=NULL)
				{ pushcallstack(curraddr);
				  setcuraddr(outhere);
				  updatewindow();
				}
			 }
			 break;
		  default:
			 break;
		}
	 }
  }
}

dsmitem *disasm::findcurrentline(void)
{ dsmitem titem,*tblock;
  lptr outhere;
  int i;
  // strategy
  // - use pointer to first item if available (so comments,etc included in list
  // - otherwise use address.
  titem.addr=curraddr;
  titem.type=subitem;
  find((listitem)&titem);
  tblock=(dsmitem *)nextiterator();
  if(tblock!=NULL)
  { if((tblock->addr.segm==curraddr.segm)&&(tblock->addr<curraddr)&&
	  (tblock->addr+tblock->length>curraddr))
	  curraddr.offs=tblock->addr.offs;
  }
  if(tblock!=NULL)
  { if(tblock->addr<curraddr)
		tblock=(dsmitem *)nextiterator();
  }
  outhere=curraddr;
  for(i=0;i<usersel;i++)
  { if(tblock!=NULL)
	 { if(outhere==tblock->addr)
		{ outhere+=tblock->length;
		  tblock=(dsmitem *)nextiterator();
		}
		else outhere++;
	 }
	 else outhere++;
	 // check if gone beyond seg, get next seg.
	 if(dta.beyondseg(outhere))
		dta.nextseg(&outhere);
	 if(!outhere.segm) break;
  }
  if(tblock!=NULL)
	 if(outhere!=tblock->addr)
		return NULL;
  return tblock;
}

void disasm::findcurrentaddr(lptr *loc)
{ dsmitem titem,*tblock;
  lptr outhere;
  int i;
  // strategy
  // - use pointer to first item if available (so comments,etc included in list
  // - otherwise use address.
  titem.addr=curraddr;
  titem.type=subitem;
  find((listitem)&titem);
  tblock=(dsmitem *)nextiterator();
  if(tblock!=NULL)
  { if((tblock->addr.segm==curraddr.segm)&&(tblock->addr<curraddr)&&
	  (tblock->addr+tblock->length>curraddr))
	  curraddr.offs=tblock->addr.offs;
  }
  if(tblock!=NULL)
  { if(tblock->addr<curraddr)
		tblock=(dsmitem *)nextiterator();
  }
  outhere=curraddr;
  for(i=0;i<usersel;i++)
  { if(tblock!=NULL)
	 { if(outhere==tblock->addr)
		{ outhere+=tblock->length;
		  tblock=(dsmitem *)nextiterator();
		}
		else outhere++;
	 }
	 else outhere++;
	 // check if gone beyond seg, get next seg.
	 if(dta.beyondseg(outhere))
		dta.nextseg(&outhere);
	 if(!outhere.segm) break;
  }
  if(outhere.segm) (*loc)=outhere;
  else (*loc)=curraddr;
}

void disasm::discomment(lptr loc,dsmitemtype typ,byte *comment)
{ dsmitem *newdsm;
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=typ;
  newdsm->override=over_null;
  newdsm->tptr=(void *)comment;
  newdsm->length=0;
  newdsm->data=comment;
  addto((listitem)newdsm);
  updatewindowifinrange(loc);
}

void disasm::delcomment(lptr loc,dsmitemtype typ)
{ dsmitem ddsm,*fdsm;
  ddsm.addr=loc;
  ddsm.type=typ;
  fdsm=(dsmitem *)find((listitem)&ddsm);
  if(fdsm!=NULL)
	 if((fdsm->addr==loc)&&(fdsm->type==typ))
	 { delfrom((listitem)fdsm);
	 }
  updatewindowifinrange(loc);
}

void disasm::outnameloc(dsmitem *inst)
{ PrintBuff("%s:",inst->data);
}

void disasm::outcomment(dsmitem *inst)
{ PrintBuff(";%s",inst->tptr);
}

void disasm::pushcallstack(lptr loc)
{ int i;
  if(stacktop==CALLSTACKSIZE)  // need to remove bottom item from stack
  { for(i=0;i<CALLSTACKSIZE-1;i++)
	 { callstack[i]=callstack[i+1];
		stacktop--;
	 }
  }
  callstack[stacktop]=loc;
  stacktop++;
}

void disasm::popcallstack(lptr *loc)
{ if(!stacktop)return;
  stacktop--;
  (*loc)=callstack[stacktop];
}

void disasm::savecuraddr(void)
{ pushcallstack(curraddr);
}

BOOL disasm::interpretmod(byte *data,dword *toffs,byte *indexreg,byte *indexreg2,byte *indexamount,int *numjumps)
{ byte rm,modrm,sib;
  rm=(byte)((data[0]&0xc0)>>6);
  modrm=(byte)(data[0]&0x07);
  switch(rm)
  { case 0:
		if(options.mode32)
		{ if(modrm==5)   // disp32 only.
		  { *toffs=((dword *)(&data[1]))[0];
			 *numjumps=1;
		  }
		  else if(modrm==4)        // case 4=sib
		  { sib=data[1];
			 if((sib&0x07)==5) *toffs=((dword *)(&data[2]))[0];  // disp32
			 else return FALSE; // no disp
			 if(((sib>>3)&0x07)==4) *numjumps=1;  // no scaled index reg
			 else
			 { *indexreg=(byte)((sib>>3)&0x07);
				switch(sib>>6)
				{ case 0:
					 *indexamount=1;
					 break;
				  case 1:
					 *indexamount=2;
					 break;
				  case 2:
					 *indexamount=4;
					 break;
				  case 3:
					 *indexamount=8;
					 break;
				}
			 }
		  }
		  else return FALSE; // no disp
		}
		else // 16-bit mode
		{ if(modrm==6) // disp16 only
		  { *toffs=((word *)(&data[1]))[0];
			 *numjumps=1;
		  }
		  else return FALSE; // no disp
		}
		break;
	 case 1:
		return FALSE; // all disp8 offsets - don't follow
	 case 2:
		if(options.mode32)
		{ if(modrm==4)        // case 4=sib
		  { sib=data[1];
			 *toffs=((dword *)(&data[2]))[0];
			 *indexreg2=(byte)(sib&0x07);
			 if(((sib>>3)&0x07)==4); // no scaled index reg
			 else
			 { *indexreg=(byte)((sib>>3)&0x07);
				switch(sib>>6)
				{ case 0:
					 *indexamount=1;
					 break;
				  case 1:
					 *indexamount=2;
					 break;
				  case 2:
					 *indexamount=4;
					 break;
				  case 3:
					 *indexamount=8;
					 break;
				}
			 }
		  }
		  else
		  { *toffs=((dword *)(&data[1]))[0];
			 *indexreg2=(byte)(data[0]&0x07);
		  }
		}
		else // 16bit mode
		{ *toffs=((word *)(&data[1]))[0];
		  *indexreg=(byte)(data[0]&0x07); // NB double index reg eg bx+si
		}
		break;
	 case 3:
		// case 3 - no jump table offset present. indirect jump.
		return FALSE;
  }
  return TRUE;
}

// this is a mess - needs rewriting badly
void disasm::disjumptable(lptr loc)
{ dsmitem findit,*investigate;
  dsegitem *dblock,*idblock;
  byte *data;
  byte pbyte;       // prefix byte
  lptr t,it,index,xr;
  int numjumps,inumjumps;
  int i;
  byte indexreg,indexamount,indexreg2;
  byte iindexreg,iindexamount,iindexreg2;
  char *tablename,tablenum[10];
  BOOL itable;
  numjumps=0;
  indexreg=0;
  indexreg2=0;
  indexamount=0;
  inumjumps=0;
  iindexreg=0;
  iindexreg2=0;
  iindexamount=0;
  findit.addr=loc;
  findit.type=dsmcode;
  investigate=(dsmitem *)find((listitem)&findit);
  if(investigate==NULL)return;
  // check that inst is still there/ correct type of jump
  // adjust for any segment overrides added since
  if((loc-investigate->addr)<4)loc.offs=investigate->addr.offs;
  if((investigate->addr!=loc)||(investigate->type!=dsmcode))
	 return;
  if(((asminstdata *)investigate->tptr)->arg1!=ARG_MODRM)
	 return;
  if(!(investigate->flags&(FLAGS_JMP|FLAGS_CALL|FLAGS_CJMP))) return;
  data=investigate->data+investigate->modrm;
  if(!interpretmod(data,&t.offs,&indexreg,&indexreg2,&indexamount,&numjumps)) return;
  // find target - jump table, need to use default ds:/ check for cs: override
  if(investigate->flags&FLAGS_SEGPREFIX)
  { pbyte=investigate->data[0];
	 if((pbyte==0x66)||(pbyte==0x67))pbyte=investigate->data[1];
	 if((pbyte==0x66)||(pbyte==0x67))pbyte=investigate->data[2];
  }
  t.segm=options.dseg;
  if(pbyte==0x2e)t.segm=loc.segm;
  dblock=dta.findseg(t);      // find segment item.
  if(dblock==NULL)return;
  // look at previous instructions for number of entries
  itable=FALSE;
  if(!numjumps)
  { for(i=0;i<10;i++)
	 { investigate=(dsmitem *)lastiterator();
		if(investigate==NULL) break;                                      // no previous insts
		if((investigate->addr.segm!=loc.segm)||(investigate->addr.offs+50<loc.offs)) break; // too far back
		if(investigate->type!=dsmcode) i--;                               // skip non-code
		else
		if(((!strcmp(((asminstdata *)investigate->tptr)->name,"mov"))
		  ||(!strcmp(((asminstdata *)investigate->tptr)->name,"movzx")))
		  &&((((asminstdata *)investigate->tptr)->arg1==ARG_MODRM)
		  ||(((asminstdata *)investigate->tptr)->arg1==ARG_MODRM8)))
		{ if(!itable)
			 if(interpretmod(investigate->data+investigate->modrm,&it.offs,&iindexreg,&iindexreg2,&iindexamount,&inumjumps))
			 { itable=TRUE;
				index.offs=investigate->addr.offs;
			 }
		}
		else
		if(((!strcmp(((asminstdata *)investigate->tptr)->name,"mov"))
		  ||(!strcmp(((asminstdata *)investigate->tptr)->name,"movzx")))
		  &&((((asminstdata *)investigate->tptr)->arg2==ARG_MODRM)
		  ||(((asminstdata *)investigate->tptr)->arg2==ARG_MODRM8)))
		{ if(!itable)
			 if(interpretmod(investigate->data+investigate->modrm,&it.offs,&iindexreg,&iindexreg2,&iindexamount,&inumjumps))
			 { itable=TRUE;
				index.offs=investigate->addr.offs;
			 }
		}
		else
		if(investigate->data[0]==0x3b)    // cmp inst
		{ if(options.mode32) numjumps=((dword *)(&investigate->data[1]))[0]+1;
		  else numjumps=((word *)(&investigate->data[1]))[0]+1;
		  break;
		}
		else
		if(investigate->data[0]==0x3d)    // cmp inst
		{ if(options.mode32) numjumps=((dword *)(&investigate->data[1]))[0]+1;
		  else numjumps=((word *)(&investigate->data[1]))[0]+1;
		  break;
		}
		else
		if((investigate->data[0]==0x83)&&(investigate->data[1]>=0xc0)) // cmp reg,imm8
		{ numjumps=investigate->data[2]+1;
		  break;
		}
	 }
  }
  if(itable)
  { it.segm=t.segm;
	 idblock=dta.findseg(it);
	 if(idblock==NULL)return;
	 inumjumps=numjumps;
	 if((inumjumps<2)||(inumjumps>0x100))return;
	 numjumps=0;
	 for(i=0;i<inumjumps;i++)
	 { if(it+i>dblock->addr+dblock->size)return;
		if((dblock->data+(it-dblock->addr)+i)[0]>numjumps)
		  numjumps=(dblock->data+(it-dblock->addr)+i)[0];
	 }
	 numjumps++;
  }
  // add code disassemblies to scheduler
  // name it
  if((!numjumps)||(numjumps>0x100))return;
  tablename=new char[20];
  if(numjumps>1)
  { jtables++;
	 wsprintf(tablenum,"%d",jtables);
	 strcpy(tablename,"jumptable_");
  }
  else
  { irefs++;
	 wsprintf(tablenum,"%d",irefs);
	 strcpy(tablename,"indirectref_");
  }
  strcat(tablename,tablenum);
  if(!name.isname(t))scheduler.addtask(namecurloc,priority_nameloc,t,tablename);
  xrefs.addxref(t,loc);
  if(itable)
  { tablename=new char[20];
	 itables++;
	 wsprintf(tablenum,"%d",itables);
	 strcpy(tablename,"indextable_");
	 strcat(tablename,tablenum);
	 if(!name.isname(it))scheduler.addtask(namecurloc,priority_nameloc,it,tablename);
	 index.segm=loc.segm;
	 xrefs.addxref(it,index);
  }
  // disassemble data
  // disassemble code
  if(!indexamount)
	 if(options.mode32) indexamount=4;
	 else indexamount=2;
  for(i=0;i<numjumps;i++)
  { if(t+i*indexamount>dblock->addr+dblock->size) return;
	 if(options.mode32)
	 { xr.assign(loc.segm,((dword *)(dblock->data+(t-dblock->addr)+i*indexamount))[0]);
		scheduler.addtask(dis_datadsoffword,priority_data,t+i*indexamount,NULL);
		scheduler.addtask(dis_code,priority_definitecode,xr,NULL);
		xrefs.addxref(xr,t+i*indexamount);
	 }
	 else
	 { xr.assign(loc.segm,((word *)(dblock->data+(t-dblock->addr)+i*indexamount))[0]);
		scheduler.addtask(dis_dataword,priority_data,t+i*indexamount,NULL);
		scheduler.addtask(dis_code,priority_definitecode,xr,NULL);
		xrefs.addxref(xr,t+i*indexamount);
	 }
  }
}

void disasm::dumptofile(char *fname,BOOL printaddrs)
{ SECURITY_ATTRIBUTES securityatt;
  HANDLE efile;
  dword num;
  dsmitem *tblock;
  dsegitem *dblock;
  lptr outhere,nptr;
  int i;
  securityatt.nLength=sizeof(SECURITY_ATTRIBUTES);
  securityatt.lpSecurityDescriptor=NULL;
  securityatt.bInheritHandle=FALSE;
  efile=CreateFile(fname,GENERIC_WRITE,0,&securityatt,CREATE_NEW,0,NULL);
  if(efile==INVALID_HANDLE_VALUE)
  { MessageBox(mainwindow,"File Creation Failed","Borg Disassembler Alert",MB_OK);
	 return;
  }
  scheduler.stopthread();
  WriteFile(efile,"; ",2,&num,NULL);
  WriteFile(efile,winname,strlen(winname),&num,NULL);
  WriteFile(efile,"\r\n;\r\n",5,&num,NULL);
  WriteFile(efile,hdr,strlen(hdr),&num,NULL);
  WriteFile(efile,hdr2,strlen(hdr2),&num,NULL);
  WriteFile(efile,"\r\n",2,&num,NULL);
  // find current position.
  resetiterator();
  tblock=(dsmitem *)nextiterator();
  // now tblock= first position
  dta.resetiterator();
  dblock=(dsegitem *)dta.nextiterator();
  outhere=dblock->addr;
  while(dblock!=NULL)
  { ClearBuff(); // clear buffer - ready to start
	 for(i=0;i<buffer_lines-1;i++)
	 { if(tblock!=NULL)
		{ if(outhere==tblock->addr)
		  { switch(tblock->type)
			 { case dsmcode:
				  outinst(tblock,printaddrs);
				  break;
				case dsmnameloc:
				  outnameloc(tblock);
				  break;
				case dsmxref:
				  PrintBuff(";XREFS");
				  break;
				default:
				  outcomment(tblock);
				  break;
			 }
			 outhere+=tblock->length;
			 tblock=(dsmitem *)nextiterator();
		  }
		  else
		  { outdb(&outhere,printaddrs);
			 outhere++;
		  }
		}
		else
		{ outdb(&outhere,printaddrs);
		  outhere++;
		}
		// check if gone beyond seg, get next seg.
		/*if(dta.beyondseg(outhere))*/     // changed build 14
		if(outhere>dblock->addr+dblock->size)
		{ dblock=(dsegitem *)dta.nextiterator();
		  if(dblock==NULL)break;
		  outhere=dblock->addr;
		}
		if(!outhere.segm) break;
	 }
	 DumpBuff(efile);
  }
  DoneBuff();
  CloseHandle(efile);
  nptr.assign(0,0);
  scheduler.continuethread();
  scheduler.addtask(windowupdate,priority_window,nptr,NULL);
}

void disasm::disxref(lptr loc)
{ dsmitem *newdsm,*chk;
  newdsm=new dsmitem;
  newdsm->addr=loc;
  newdsm->type=dsmxref;
  newdsm->override=over_null;
  newdsm->tptr=NULL;
  newdsm->length=0;
  newdsm->data=NULL;
  chk=(dsmitem *)find((listitem)newdsm);
  if(chk!=NULL)
  { if((chk->type==dsmxref)&&(chk->addr==loc))
	 { delete newdsm;
	 }
	 else if((chk->length)&&(chk->addr.segm==loc.segm)&&(chk->addr<loc)&&(chk->addr+chk->length>loc))
	 { delete newdsm;
	 }
	 else addto((listitem)newdsm);
  }
  else addto((listitem)newdsm);
  updatewindowifinrange(loc);
}

// used by search engine
int disasm::getlength(lptr loc)
{ dsmitem tmp,*fnd;
  tmp.addr=loc;
  tmp.type=dsmcode;
  fnd=(dsmitem *)find((listitem)&tmp);
  if(fnd==NULL)return 1;
  if((fnd->addr!=loc)||(fnd->type!=dsmcode))return 1;
  return fnd->length;
}

BOOL disasm::savedb(HANDLE sfile,byte *filebuff)
{ dword ndsms,num;
  dsmitemsave structsave;
  dsmitem *currdsm;
  ndsms=numlistitems();
  resetiterator();
  if(!WriteFile(sfile,&ndsms,sizeof(dword),&num,NULL))return FALSE;
  while(ndsms)
  { currdsm=(dsmitem *)nextiterator();
	 structsave.addr=currdsm->addr;
	 structsave.type=currdsm->type;
	 structsave.length=currdsm->length;
	 structsave.modrm=currdsm->modrm;
	 structsave.mode32=currdsm->mode32;
	 structsave.override=currdsm->override;
	 structsave.flags=currdsm->flags;
	 if(structsave.type==dsmxref)
	 { structsave.fileoffset=0;  // NULL ptrs
		structsave.tptroffset=0;
	 }
	 else if(structsave.type==dsmcode)
	 { structsave.fileoffset=currdsm->data-filebuff;
		structsave.tptroffset=((asminstdata *)currdsm->tptr)->uniqueid;
	 }
	 else
	 { structsave.fileoffset=strlen((char *)currdsm->data)+1; // strlen
		structsave.tptroffset=0; // points to str as well
	 }
	 if(!WriteFile(sfile,&structsave,sizeof(dsmitemsave),&num,NULL))return FALSE;
	 if((structsave.type!=dsmxref)&&(structsave.type!=dsmcode))
	 { if(!WriteFile(sfile,currdsm->tptr,structsave.fileoffset,&num,NULL))return FALSE;
	 }
	 ndsms--;
  }
  // need to save callstack and some other stuff too.
  if(!WriteFile(sfile,&curraddr,sizeof(lptr),&num,NULL))return FALSE;
  if(!WriteFile(sfile,&subitem,sizeof(dsmitemtype),&num,NULL))return FALSE;
  if(!WriteFile(sfile,&stacktop,sizeof(int),&num,NULL))return FALSE;
  if(!WriteFile(sfile,&itables,sizeof(int),&num,NULL))return FALSE;
  if(!WriteFile(sfile,&jtables,sizeof(int),&num,NULL))return FALSE;
  if(!WriteFile(sfile,&irefs,sizeof(int),&num,NULL))return FALSE;
  if(!WriteFile(sfile,callstack,sizeof(lptr)*CALLSTACKSIZE,&num,NULL))return FALSE;
  return TRUE;
}

BOOL disasm::loaddb(HANDLE sfile,byte *filebuff)
{ dword ndsms,num;
  dsmitemsave structsave;
  dsmitem *currdsm;
  int asminstctr;
  asminstdata *findasm;
  if(!ReadFile(sfile,&ndsms,sizeof(dword),&num,NULL))return FALSE;
  while(ndsms)
  { currdsm=new dsmitem;
	 if(!ReadFile(sfile,&structsave,sizeof(dsmitemsave),&num,NULL))return FALSE;
	 currdsm->addr=structsave.addr;
	 currdsm->type=structsave.type;
	 currdsm->length=structsave.length;
	 currdsm->modrm=structsave.modrm;
	 currdsm->mode32=structsave.mode32;
	 currdsm->override=structsave.override;
	 currdsm->flags=structsave.flags;
	 if(structsave.type==dsmxref)
	 { currdsm->data=NULL;
		currdsm->tptr=NULL;
	 }
	 else if(structsave.type==dsmcode)
	 { currdsm->data=structsave.fileoffset+filebuff;
		// now reset the tptr = asminstdata ptr (need to find it from the uniqueid)
		asminstctr=0;
		findasm=reconstruct[asminstctr];
		while((dword)(findasm[0].uniqueid/1000)!=(dword)(structsave.tptroffset/1000))
		{ asminstctr++;
		  findasm=reconstruct[asminstctr];
		  if(findasm==NULL)
		  {
#ifdef DEBUG
			 DebugMessage("File Loader:Failed to find instruction table for %lu",structsave.tptroffset);
#endif
			 return FALSE;
		  }
		}
		asminstctr=0;
		while(findasm[asminstctr].uniqueid!=structsave.tptroffset)
		{ asminstctr++;
		  if((!findasm[asminstctr].instbyte)&&(!findasm[asminstctr].processor))
		  {
#ifdef DEBUG
			 DebugMessage("File Loader:Failed to find instruction %lu",structsave.tptroffset);
#endif
			 return FALSE;
        } 
		}
		currdsm->tptr=(void *)&findasm[asminstctr];
	 }
	 else
	 { currdsm->data=new byte[structsave.fileoffset];
		currdsm->tptr=currdsm->data;
	 }
	 if((structsave.type!=dsmxref)&&(structsave.type!=dsmcode))
	 { if(!ReadFile(sfile,currdsm->tptr,structsave.fileoffset,&num,NULL))return FALSE;
	 }
	 addto((listitem)currdsm);
	 ndsms--;
  }
  // need to save callstack and some other stuff too.
  if(!ReadFile(sfile,&curraddr,sizeof(lptr),&num,NULL))return FALSE;
  if(!ReadFile(sfile,&subitem,sizeof(dsmitemtype),&num,NULL))return FALSE;
  if(!ReadFile(sfile,&stacktop,sizeof(int),&num,NULL))return FALSE;
  if(!ReadFile(sfile,&itables,sizeof(int),&num,NULL))return FALSE;
  if(!ReadFile(sfile,&jtables,sizeof(int),&num,NULL))return FALSE;
  if(!ReadFile(sfile,&irefs,sizeof(int),&num,NULL))return FALSE;
  if(!ReadFile(sfile,callstack,sizeof(lptr)*CALLSTACKSIZE,&num,NULL))return FALSE;
  dissettable();
  setcuraddr(curraddr);
  return TRUE;
}


