//						disasm.h
//

#ifndef disasm_h
#define disasm_h

#include "list.h"
#include "common.h"
#include "proctab.h"

#define BYTEPOS 14
#define ASMPOS 31
#define ARGPOS 38
#define COMMENTPOS 50

#define CALLSTACKSIZE 100

enum dsmitemtype {dsmnull=0,dsmsegheader,dsmnameloc=5,dsmcomment,dsmxref,dsmcode};
enum byteoverride {over_null=0,over_decimal,over_char,over_dsoffset};

struct dsmitem
{ lptr addr;
  void *tptr;                     // string pointer for comment
											 // asminstdata ptr for code
  dsmitemtype type;
  word length;
  byte modrm;              // length, offset to  modrm byte
  BOOL mode32;
  byte *data;                     // data ptr
  byteoverride override;
  dword flags;
};

struct dsmitemsave
{ lptr addr;
  dword tptroffset;
  dsmitemtype type;
  word length;
  byte modrm;
  BOOL mode32;
  dword fileoffset;
  byteoverride override;
  dword flags;
};

class disasm: private slist
{ private:
	 asmtable *itable;
	 dsmitem *lastdis;
	 lptr curraddr,outend;
	 dsmitemtype subitem;  // subindex to top line of output
	 lptr callstack[CALLSTACKSIZE];
	 int stacktop;
    int itables,jtables,irefs; // number of jump tables detected

  public:
	 disasm();
	 ~disasm();
	 void dissettable(void);
	 void discomment(lptr loc,dsmitemtype typ,byte *comment);
	 void delcomment(lptr loc,dsmitemtype typ);
	 void disblock(lptr loc);
	 void disxref(lptr loc);
	 void disjumptable(lptr loc);
	 void disdataword(lptr loc);
    void disdatadword(lptr loc);
    void disdatadsoffword(lptr loc);
	 void disdatastring(lptr loc);
	 void disdatapstring(lptr loc);
	 void disdataucstring(lptr loc);
	 void disdataupstring(lptr loc);
	 void disdatadosstring(lptr loc);
	 void disdatageneralstring(lptr loc);
	 void disargoverdec(lptr loc);
	 void disargoverhex(lptr loc);
	 void disargoverchar(lptr loc);
	 void disargoveroffsetdseg(lptr loc);
	 BOOL oktoname(lptr loc);
	 void updatewindow(void);
	 void scroller(dword amount);
	 void setcuraddr(lptr loc);
	 void savecuraddr(void);
	 void makecode(void);
	 void makeword(void);
	 void makedword(void);
	 void makestring(void);
	 void pascalstring(void);
	 void ucstring(void);
	 void upstring(void);
	 void dosstring(void);
	 void generalstring(void);
	 void argoverdec(void);
	 void argoverhex(void);
	 void argoverchar(void);
    void argoveroffsetdseg(void);
	 void undefineline(void);
	 void undefinelines(void);
    void setpos(int ypos);
	 void jumpto(void);
    void jumpback(void);
	 void findcurrentaddr(lptr *loc);
	 void findcurrentaddr2(lptr *loc);
	 void dumptofile(char *fname,BOOL printaddrs);
	 int getlength(lptr loc);
	 BOOL savedb(HANDLE sfile,byte *filebuff);
	 BOOL loaddb(HANDLE sfile,byte *filebuff);

  private:
	 dsmitem *decodeinst(byte *ibuff,byte *mcode,lptr loc,byte tabtype,BOOL omode32,int depth);
	 byte arglength(argtype a,byte modrmbyte,byte sibbyte,dword flgs,BOOL omode32);
	 void outinst(dsmitem *inst,BOOL printaddrs);
	 void outcomment(dsmitem *inst);
	 void outnameloc(dsmitem *inst);
	 void outargs(dsmitem *inst,argtype a);
	 void outdb(lptr *lp,BOOL printaddrs);
	 void outprefix(byte prefixbyte);
	 dsmitem *findcurrentline(void);
	 void updatewindowifinrange(lptr loc);
	 BOOL checkvalid(dsmitem *newdsm);
	 void cleanstring(char *str);
	 void popcallstack(lptr *loc);
	 void pushcallstack(lptr loc);
	 BOOL interpretmod(byte *data,dword *toffs,byte *indexreg,byte *indexreg2,byte *indexamount,int *numjumps);
};

extern disasm dsm;

#endif