//					list.cpp
//
// basic list class, use with inheritance.
// - maintains a list class of pointers.
// - can add/insert items
// - can use binary search
// - will resize array upwards in 4k chunks.
// - downsizing has not been added but will be carried out when a disassembly
// is saved and reloaded as a database.

#include <windows.h>
#include "list.h"

slist::slist(void)
{ listptr=new listitem[1024];
  listiter=0;
  listsize=0;
  maxsize=1024;
  cfunc=NULL;
  delfunc=NULL;
}

slist::~slist(void)
{ dword i=0;
  while(i<listsize)
  { delfunc(listptr[i]);
	 i++;
  }
  delete listptr;
}

void slist::setcomparefunc(int (*vfunc)(listitem,listitem))
{ cfunc=vfunc;
}

void slist::setdelfunc(void (*dfunc)(listitem))
{ delfunc=dfunc;
}

// add an item to the list
void slist::addto(listitem ptr)
{ dword i;
  listitem *nlist;
  if(listsize==maxsize)
  { // resize list
	 maxsize+=1024;
	 nlist=new listitem[maxsize];
	 for(i=0;i<listsize;i++) nlist[i]=listptr[i];
	 delete listptr;
	 listptr=nlist;
  }
  if(find(ptr)==NULL)
  { // empty list
	 listptr[0]=ptr;
	 listsize=1;
	 return;
  }
  else
  { // use listiter set by find
	 // ensure can add to end
	 if(cfunc(listptr[listiter],ptr)==-1) listiter++;
	 // move array up
	 for(i=listsize;i>listiter;i--) listptr[i]=listptr[i-1];
	 listptr[listiter]=ptr;
	 listsize++;
  }
}

// delete an item from the list
void slist::delfrom(listitem ptr)
{ dword i;
  if(!listsize) return;
  if(find(ptr)==NULL)
  { // empty list
	 return;
  }
  if(!cfunc(listptr[listiter],ptr))  // ensure equal
  { delfunc(listptr[listiter]);
	 // move the rest down
	 for(i=listiter;i<listsize-1;i++) listptr[i]=listptr[i+1];
	 listsize--;
  }
}

// finds ptr in the list, using a binary search
// returns a pointer to the nearest item
// sets list iterator to that item
listitem slist::find(listitem ptr)
{ dword i,j,k;
  if(!listsize) return NULL;
  i=0;
  j=listsize-1;
  while(i!=j)
  { k=(i+j)>>1;
	 switch(cfunc(listptr[k],ptr))
	 { case -1:      // listptr[k]->cmp < ptr->cmp
		  if(j-i==1)  // special case
		  { if(i==k)
			 { if(cfunc(listptr[j],ptr)!=1) i=j;
				else j=i;
			 }
			 else i=j;
		  }
		  else i=k;   // move lower bound up
		  break;
		case 0:       // listptr[k]->cmp==ptr->cmp
		  j=k;        // only gets the first one of this type
		  break;
		case 1:       // listptr[k]->cmp > ptr->cmp
		  if(j-i==1)  // special case
		  { if(i==k) j=k;
			 else j=i;
		  }
		  else j=k;   // move upper bound down
		  break;
	 }
  }
  listiter=i;
  return listptr[i];
}

// resets list iterator to start of list
void slist::resetiterator(void)
{ listiter=0;
}

// obtains next item in the list, using the iterator
listitem slist::nextiterator(void)
{ if(listiter>=listsize)
  { listiter=0;
	 return NULL;
  }
  return listptr[listiter++];
}

// obtains previous item in the list, using the iterator
listitem slist::lastiterator(void)
{ if(!listiter)
  { return NULL;
  }
  listiter--;
  return listptr[listiter];
}

//returns the first item from the list and removes it from the list
listitem slist::processqueue(void)
{ dword i;
  listitem t;
  if(!listsize)return NULL;
  t=listptr[0];
  for(i=0;i<listsize-1;i++)
	 listptr[i]=listptr[i+1];
  listsize--;
  return t;
}

//returns the first item from the list without removing it
listitem slist::peekfirst(void)
{ if(!listsize)return NULL;
  return listptr[0];
}

dword slist::numlistitems(void)
{ return listsize;
}


