//				mainwind.cpp

#include <windows.h>
#include "mainwind.h"
#include "dasm.h"

// this file controls the output to the main window.
// it works like this:
//   a small buffer of output lines is kept.
//   this is used to repaint the screen when needed.
//   the main thread can update this buffer by calling the various routines.
//   if the user moves in the window then a schedule message is sent to repaint,
//   and this is processed as soon as possible.

UINT lastline=0;
volatile BOOL bufferready=TRUE;
char MainBuff[buffer_lines*max_length+1000]; // lets hope it doesn't overflow
                                             // wvsprintf sucks!
UINT nScreenRows;
UINT usersel;  // user selection. (line number).
int rrr=0,cyc;
int hpos;
BOOL userselonscreen=FALSE;
HPEN greenline;

void horizscroll(int amount)
{ EnterCriticalSection(&cs);
  hpos+=amount;
  if(hpos<0)hpos=0;  // max-size checked in dopaint.
  LeaveCriticalSection(&cs);
  InvalidateRect(mainwindow,NULL,TRUE);
}

// should be called before each repainting of the buffer.
void ClearBuff(void)
{ int i;
  EnterCriticalSection(&cs);
  for(i=0;i<buffer_lines;i++)
  { MainBuff[i*max_length]=0;
  }
  lastline=0;
  bufferready=FALSE;
  LeaveCriticalSection(&cs);
}

// should call this when the thread has finished filling the buffer.
void DoneBuff(void)
{ EnterCriticalSection(&cs);
  bufferready=TRUE;
  LeaveCriticalSection(&cs);
}

// adds next line to the buffer.
void PrintBuff(char *szFormat,...)
{ EnterCriticalSection(&cs);
  va_list vaArgs;
  va_start(vaArgs,szFormat);
  if(lastline<buffer_lines-1)
  { wvsprintf(&MainBuff[lastline*max_length],szFormat,vaArgs);
	 lastline++;
  }
  va_end(vaArgs);
  LeaveCriticalSection(&cs);
}

// adds blanks to previous line to set position
void LastPrintBuffEpos(UINT xpos)
{ UINT i;
  int spos;
  EnterCriticalSection(&cs);
  if(lastline)
  { spos=(lastline-1)*max_length;
	 i=strlen(&MainBuff[spos]);
	 while(i<xpos)
	 { MainBuff[spos+i]=' ';
		MainBuff[spos+i+1]=0;
		i++;
	 }
  }
  LeaveCriticalSection(&cs);
}

// add to last line
void LastPrintBuff(char *szFormat,...)
{ int spos;
  EnterCriticalSection(&cs);
  va_list vaArgs;
  va_start(vaArgs,szFormat);
  if(lastline)
  { spos=(lastline-1)*max_length;
	 wvsprintf(&MainBuff[spos+strlen(&MainBuff[spos])],szFormat,vaArgs);
  }
  va_end(vaArgs);
  LeaveCriticalSection(&cs);
}

void DumpBuff(HANDLE efile)
{ dword num;
  int i;
  for(i=0;i<lastline;i++)
  { WriteFile(efile,&MainBuff[i*max_length],lstrlen(&MainBuff[i*max_length]),&num,NULL);
	 WriteFile(efile,"\r\n",2,&num,NULL);
  }
}

// mainwindow paint routine
//
// anyone know how to reduce flicker here ?
// - perhaps erase and draw a line at a time ?
// - help!!
void DoPaint(HWND hWnd,int cxChar,int cyChar)
{ HDC         hDC;   // handle for the display device
  PAINTSTRUCT ps;    // holds PAINT information
  UINT        nI;
  HGDIOBJ pobj;
  RECT rRect;
  int startpt;

  cyc=cyChar;
  while(!bufferready)Sleep(0);                   // wait if filling buffer
  if(KillThread)return;
  EnterCriticalSection(&cs);
  memset(&ps, 0x00, sizeof(PAINTSTRUCT));
  hDC = BeginPaint(hWnd, &ps);
  // Included as the background is not a pure color
  SetBkMode(hDC,TRANSPARENT);
  SelectObject(hDC,GetStockObject(SYSTEM_FIXED_FONT));
  GetClientRect(hWnd,&rRect);
  if(hpos>max_length-(rRect.right/cxChar))hpos=max_length-rRect.right/cxChar;
  if(rrr!=rRect.right)
  { rrr=rRect.right;
	 SetScrollRange(hWnd,SB_HORZ,0,max_length-rRect.right/cxChar,TRUE);
  }
  SetScrollPos(hWnd,SB_HORZ,hpos,TRUE);
  nScreenRows = rRect.bottom/cyChar;
  ShowScrollBar (hWnd, SB_VERT, TRUE);
  ShowScrollBar (hWnd, SB_HORZ, max_length>(rRect.right/cxChar));
  startpt=0;
  if(userselonscreen)
  { pobj=SelectObject(hDC,greenline);
	 MoveToEx(hDC,0,usersel*cyChar+cyChar/2+1,NULL);
	 LineTo(hDC,rRect.right,usersel*cyChar+cyChar/2+1);
	 DeleteObject(pobj);
  }
  for(nI=startpt;nI<lastline;nI++)
  { TabbedTextOut(hDC,2-hpos*cxChar,nI*cyChar,&MainBuff[nI*max_length],
		lstrlen(&MainBuff[nI*max_length]),0,NULL,2-hpos*cxChar);
  }
  // Inform Windows painting is complete
  EndPaint(hWnd,&ps);
  LeaveCriticalSection(&cs);
}

void InitPaintSubsystem(int cyChar)
{ greenline=CreatePen(PS_SOLID,cyChar,0x0000ff00);
  hpos=0;
}


