//						relocs.cpp
//
// This is a class which maintains a list of relocation items from the exe file.
// - ie items which have been or should be relocated when the exe is loaded.
// - includes functions to load and save the table as part of the database load/save.
// - also a function to relocate the image in memory when the database and file are
//   reloaded.

#include <windows.h>

#include "relocs.h"
#include "data.h"
#include "debug.h"
#include "dasm.h"

int relocscfunc(listitem i,listitem j);
void relocsdfunc(listitem i);

// sets up compare and deletion functions for the relocs list.
// similar to many other constructors.
relocs::relocs()
{ setcomparefunc(relocscfunc);
  setdelfunc(relocsdfunc);
}

relocs::~relocs()
{
}

// adds a reloc item to the list of relocs.
void relocs::addreloc(lptr loc,reloctype type)
{ relocitem *newname;
  newname=new relocitem;
  newname->addr=loc;
  newname->type=type;
  addto((listitem)newname);
#ifdef DEBUG
  DebugMessage("Added Reloc : %04lx:%04lx",loc.segm,loc.offs);
#endif
}

// returns true if loc is a relocation item. ie if the address loc is in the reloc table.
BOOL relocs::isreloc(lptr loc)
{ relocitem checkit,*findit;
  checkit.addr=loc;
  findit=(relocitem *)find((listitem)&checkit);
  if(findit!=NULL) if(findit->addr==loc) return TRUE;
  return FALSE;
}

//compare function to order in list
int relocscfunc(listitem i,listitem j)
{ if(((relocitem *)i)->addr==((relocitem *)j)->addr)
	 return 0;
  if(((relocitem *)i)->addr>((relocitem *)j)->addr)
	 return 1;
  return -1;
}

//deletion function for list
void relocsdfunc(listitem i)
{ delete (relocitem *)i;
}

// relocs::saving to database. Saves reloc table to database file.
BOOL relocs::savedb(HANDLE sfile)
{ dword nrels,num;
  relocitem *currrel;
  nrels=numlistitems();
  resetiterator();
  // save number of reloc items
  if(!WriteFile(sfile,&nrels,sizeof(dword),&num,NULL))return FALSE;
  while(nrels)
  { currrel=(relocitem *)nextiterator();
	 // save each reloc item
	 if(!WriteFile(sfile,currrel,sizeof(relocitem),&num,NULL))return FALSE;
	 nrels--;
  }
  return TRUE;
}

// relocs::loading from database. Loads the reloc table from the database file.
// reconstructs the list of reloc items.
BOOL relocs::loaddb(HANDLE sfile)
{ dword nrels,num;
  relocitem *currrel;
  // get number of items
  if(!ReadFile(sfile,&nrels,sizeof(dword),&num,NULL))return FALSE;
  while(nrels)
  { currrel=new relocitem;
	 // load each and add to list
	 if(!ReadFile(sfile,currrel,sizeof(relocitem),&num,NULL))return FALSE;
	 addto((listitem)currrel);
	 nrels--;
  }
  return TRUE;
}

// this is used after loading a database. It relocates the file again, without
// doing this from scratch on loading the file image.
BOOL relocs::relocfile(void)
{ dsegitem *ds;
  relocitem *ri;
  resetiterator();
  ri=(relocitem *)nextiterator();
  while(ri!=NULL)
  { // relocate item.
	 ds=dta.findseg(ri->addr);
	 if(ds!=NULL)   // changed in build 14, used to return FALSE if not found
	 { switch(ri->type)
		{ case RELOC_NONE:
			 break;
		  case RELOC_SEG:
			 ((word *)(&(ds->data[ri->addr.offs-ds->addr.offs])))[0]+=options.loadaddr.segm;
			 break;
		  case RELOC_OFFS16:
			 break;
		  case RELOC_OFFS32:
			 break;
		  case RELOC_SEGOFFS16:
			 break;
		  case RELOC_SEGOFFS32:
			 break;
		  default:
			 return FALSE;
		}
	 }
	 ri=(relocitem *)nextiterator();
  }
  return TRUE;
}
