//				schedule.cpp
//
// main task scheduler for the program
// - maintains a task list of things to be done

#include <windows.h>
#include "schedule.h"
#include "disasm.h"
#include "data.h"
#include "names.h"
#include "imports.h"
#include "exports.h"
#include "dasm.h"
#include "debug.h"
#include "mainwind.h"

int taskcfunc(listitem i,listitem j);
void taskdfunc(listitem i);

schedule::schedule()
{ setcomparefunc(taskcfunc);
  setdelfunc(taskdfunc);
  threadpause=FALSE;
  threadstopped=FALSE;
}

schedule::~schedule()
{
}

//adds a task to the schedule list, using the priority to sort it
void schedule::addtask(tasktype ttype,priority p,lptr loc,char *comment)
{ taskitem *titem;
  static dword tnumt=0;
  // limit window updates added to queue - called be called many times by disasm.
  if(ttype==windowupdate)
  { EnterCriticalSection(&cs);
	 titem=(taskitem *)peekfirst();
	 LeaveCriticalSection(&cs);
	 if(titem!=NULL)
		if(titem->ttype==windowupdate) return;
  }
  titem=new taskitem;
  titem->ttype=ttype;
  titem->p=p;
  titem->addr=loc;
  titem->comment=comment;
  titem->tnum=tnumt;
  tnumt++;
  EnterCriticalSection(&cs);
  addto((listitem)titem);
  LeaveCriticalSection(&cs);
}

//compare function for list - uses priority to sort list
int taskcfunc(listitem i,listitem j)
{ if(((taskitem *)i)->p==((taskitem *)j)->p)
  { if(((taskitem *)i)->tnum==((taskitem *)j)->tnum)return 0;
	 if(((taskitem *)i)->tnum>((taskitem *)j)->tnum)
		return 1;
	 return -1;
  }
  if(((taskitem *)i)->p>((taskitem *)j)->p)
	 return 1;
  return -1;
}

//deletion function for list
void taskdfunc(listitem i)
{ delete(taskitem *)i;
}

//goes through the list and processes items one by one
// treats it as a queue so items are processed in priority order
// note - may need to be careful adding user items with critical sections,etc
BOOL schedule::process(void)
{ taskitem *task;
  BOOL done;
  BOOL procdany;
  done=FALSE;
  procdany=FALSE;
  // appears twice for when there are no items to process
  while(threadpause)
  { threadstopped=TRUE;
	 EnterCriticalSection(&cs);
	 task=(taskitem *)peekfirst();
	 LeaveCriticalSection(&cs);
	 if(task!=NULL)
		if(task->p<priority_userrequest)
		  break;
	 Sleep(0);
  }
  threadstopped=FALSE;
  do
  { EnterCriticalSection(&cs);
	 task=(taskitem *)peekfirst();
	 LeaveCriticalSection(&cs);
	 if(task!=NULL)
		if(task->p>priority_userrequest)
		{ while(threadpause)
		  { threadstopped=TRUE;
			 EnterCriticalSection(&cs);
			 task=(taskitem *)peekfirst();
			 LeaveCriticalSection(&cs);
			 if(task!=NULL)
				if(task->p<priority_userrequest)
				  break;
			 Sleep(0);
		  }
		}
	 threadstopped=FALSE;
	 EnterCriticalSection(&cs);
	 task=(taskitem *)processqueue();
	 LeaveCriticalSection(&cs);
	 if(task==NULL) done=TRUE;
	 else
	 { StatusMessageNItems(numlistitems());
		procdany=TRUE;
		switch(task->ttype)
		{ case dis_code:
#ifdef DEBUG
			 DebugMessage("Scheduler:Disassembling Code at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disblock(task->addr);
			 break;
		  case user_makecode:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Make Code");
#endif
			 dsm.makecode();
			 break;
		  case user_undefineline:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Undefine Line");
#endif
			 dsm.undefineline();
			 break;
		  case user_undefinelines:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Undefine Lines");
#endif
			 dsm.undefinelines();
			 break;
		  case user_jumpback:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Jump Back");
#endif
			 dsm.jumpback();
			 break;
		  case user_jumpto:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Jump To");
#endif
			 dsm.jumpto();
			 break;
		  case user_jumptoaddr:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Jump To Addr");
#endif
			 dsm.savecuraddr();
			 dsm.setcuraddr(task->addr);
			 break;
		  case windowupdate:
			 dsm.updatewindow();
			 break;
		  case scrolling:
			 dsm.scroller(task->addr.offs);
			 break;
		  case hscroll:                 // amount
			 horizscroll(task->addr.offs);
			 break;
		  case user_makedword:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Make dword");
#endif
			 dsm.makedword();
			 break;
		  case user_makeword:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Make Word");
#endif
			 dsm.makeword();
			 break;
		  case user_makestring:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Make String");
#endif
			 dsm.makestring();
			 break;
		  case user_pascalstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Pascal String");
#endif
			 dsm.pascalstring();
			 break;
		  case user_ucstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Unicode C String");
#endif
			 dsm.ucstring();
			 break;
		  case user_upstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Unicode Pascal String");
#endif
			 dsm.upstring();
			 break;
		  case user_dosstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:User DOS String");
#endif
			 dsm.dosstring();
			 break;
		  case user_generalstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:User General String");
#endif
			 dsm.generalstring();
			 break;
		  case user_argoverdec:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Argument Override Decimal");
#endif
			 dsm.argoverdec();
			 break;
		  case user_argoverhex:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Argument Override Hex");
#endif
			 dsm.argoverhex();
			 break;
		  case user_argoverchar:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Argument Override Char");
#endif
			 dsm.argoverchar();
			 break;
		  case user_argoveroffsetdseg:
#ifdef DEBUG
			 DebugMessage("Scheduler:User Argument Override Offset Dseg");
#endif
			 dsm.argoveroffsetdseg();
			 break;
		  case dis_dataword:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making word at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdataword(task->addr);
			 break;
		  case dis_datadword:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making dword at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdatadword(task->addr);
			 break;
		  case dis_datadsoffword:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making dword at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdatadsoffword(task->addr);
			 break;
		  case dis_datastring:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making string at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdatastring(task->addr);
			 break;
		  case dis_datapstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making Pascal string at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdatapstring(task->addr);
			 break;
		  case dis_dataucstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making unicode c string at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdataucstring(task->addr);
			 break;
		  case dis_dataupstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making unicode pascal string at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdataupstring(task->addr);
			 break;
		  case dis_datadosstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making DOS string at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdatadosstring(task->addr);
			 break;
		  case dis_datageneralstring:
#ifdef DEBUG
			 DebugMessage("Scheduler:Making general string at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disdatageneralstring(task->addr);
			 break;
		  case dis_argoverdec:
#ifdef DEBUG
			 DebugMessage("Scheduler:Argument Override Decimal at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disargoverdec(task->addr);
			 break;
		  case dis_argoverhex:
#ifdef DEBUG
			 DebugMessage("Scheduler:Argument Override Hex at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disargoverhex(task->addr);
			 break;
		  case dis_argoverchar:
#ifdef DEBUG
			 DebugMessage("Scheduler:Argument Override Char at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disargoverchar(task->addr);
			 break;
		  case dis_argoveroffsetdseg:
#ifdef DEBUG
			 DebugMessage("Scheduler:Argument Override Offset Dseg at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disargoveroffsetdseg(task->addr);
			 break;
		  case nameloc:
#ifdef DEBUG
			 DebugMessage("Scheduler:Naming : %04lx:%04lx to %s",task->addr.segm,task->addr.offs,task->comment);
#endif
			 name.addname(task->addr,task->comment);
			 break;
		  case dis_xref:
#ifdef DEBUG
			 DebugMessage("Scheduler:Adding Xref at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disxref(task->addr);
			 break;
		  case namecurloc:
#ifdef DEBUG
			 DebugMessage("Scheduler:Naming : %04lx:%04lx to %s",task->addr.segm,task->addr.offs,task->comment);
#endif
			 if((!task->addr.segm)&&(!task->addr.offs))name.namecurrloc(task->comment);
			 else name.addname(task->addr,task->comment);
			 delete task->comment;
			 break;
		  case dis_segheader:
#ifdef DEBUG
			 DebugMessage("Scheduler:Segheader at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dta.segheader(task->addr);
			 break;
		  case dis_jumptable:
#ifdef DEBUG
			 DebugMessage("Scheduler:Disassembling Jumptable at : %04lx:%04lx",task->addr.segm,task->addr.offs);
#endif
			 dsm.disjumptable(task->addr);
			 break;
		  case dis_ordimport:
#ifdef DEBUG
			 DebugMessage("Scheduler:Import : %04lx:%04lx to %s",task->addr.segm,task->addr.offs,task->comment);
#endif
			 import.addname(task->addr,task->comment);
			 delete task->comment;
			 break;
		  case dis_import:
#ifdef DEBUG
			 DebugMessage("Scheduler:Import : %04lx:%04lx to %s",task->addr.segm,task->addr.offs,task->comment);
#endif
			 import.addname(task->addr,task->comment);
			 break;
		  case dis_ordexport:
#ifdef DEBUG
			 DebugMessage("Scheduler:Export : %04lx:%04lx to %s",task->addr.segm,task->addr.offs,task->comment);
#endif
			 export.addname(task->addr,task->comment);
			 delete task->comment;
			 break;
		  case dis_export:
#ifdef DEBUG
			 DebugMessage("Scheduler:Export : %04lx:%04lx to %s",task->addr.segm,task->addr.offs,task->comment);
#endif
			 export.addname(task->addr,task->comment);
			 break;
		  case quitborg:
          return TRUE;
		  default:
			 break;
		}
		delete task;
	 }
	 if(KillThread)return TRUE;
  } while(!done);
  return procdany;
}

dword schedule::sizelist(void)
{ return numlistitems();
}

void schedule::stopthread(void)
{ while(!threadstopped)
  { threadpause=TRUE;
	 Sleep(0);
  }
}

void schedule::continuethread(void)
{ while(threadstopped)
  { threadpause=FALSE;
	 Sleep(0);
  }
}

